'''Auton Labs implementation of Gaussian Mixture Models for distributed data.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pickle
from typing import Dict, Iterator, List, Optional, Tuple

import numpy as np
import pandas as pd
from scipy.linalg import solve_triangular  # type: ignore[import]
from scipy.optimize import Bounds, minimize  # type: ignore[import]
from scipy.special import logsumexp  # type: ignore[import]


from ...catalog.catalog import upcast
from ...config_components.distributed_config import DistributedConfig
from ...problem_def.task import DataType, TaskType
from ...wrangler.constants import Defaults  # type: ignore[import]
from ...wrangler.dataset import Dataset, Metadata
from ...wrangler.logger import Logger

from ..impl.distributed_algorithm_instance import DistributedAlgorithmInstance, NeighborState
from ..impl.algorithm import Algorithm, HyperparamError, MemoryAlgorithmCatalog


logger = Logger(__file__, to_stdout=False).logger()


class AutonGaussianMixtureModelsNeighbor(NeighborState):  # pylint: disable=too-many-public-methods

    '''State of a neighborfor Gaussian Mixture Models.'''
    # TODO(Piggy): Consider extracting the state manipulation methods into a separate type.
    # This would solve the problem of the return type of the grad_* functions, and the
    # too-man-public-methods warning.
    _K: int  # pylint: disable=invalid-name

    _full_covariance: bool = True
    _mix: np.ndarray
    _means: np.ndarray
    _cov: np.ndarray
    _cov_tril: np.ndarray

    def __init__(self,
                 mix: np.ndarray,
                 means: np.ndarray,
                 cov_tril: np.ndarray,
                 full: bool,
                 K: Optional[int] = None,  # pylint: disable=invalid-name,
                 columns: Optional[List[str]] = None
                 ):
        super().__init__(columns=columns)
        if K is None:
            K = mix.shape[0]
        self._mix = mix
        self._means = means
        self._cov_tril = cov_tril
        self._full_covariance = full
        self._K = K  # pylint: disable=invalid-name

    def __str__(self):
        return (f'AutonGaussianMixtureModelsNeighbor(K={self._K}, '
                f'mix shape: {self._mix.shape}, '
                f'means shape: {self._means.shape}, '
                f'cov_tril shape: {self._cov_tril.shape}, '
                f'full: {self._full_covariance})')

    def __repr__(self):
        return (f'AutonGaussianMixtureModelsNeighbor(K={self._K},\n'
                f'mix={self._mix},\n'
                f'means={self._means},\n'
                f'cov_tril={self._cov_tril},\n'
                f'full={self._full_covariance})')

    def encode(self) -> bytes:
        '''Encode the state of the neighbor.'''
        if self._full_covariance:
            _, d = self._means.shape
            cov = np.stack(
                [self._cov_tril[k][np.tril_indices(d)] for k in range(self._K)],
                0)
        else:
            cov = self._cov_tril
        retval = pickle.dumps((self._mix, self._means, cov, self.columns))
        return retval

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'AutonGaussianMixtureModelsNeighbor':
        '''Decode the state of the neighbor.'''
        mix, means, cov, cols = pickle.loads(serialized_model)
        # TODO(Merritt): add some asserts here
        K, d = means.shape  # pylint: disable=invalid-name
        if cov.shape[1] > means.shape[1]:
            full_covariance = True
            cov_tril = np.zeros((K, d, d))
            for k in range(K):
                cov_tril[k][np.tril_indices(d)] = cov[k]
        else:
            full_covariance = False
            cov_tril = cov
        if cols is not None:
            # allow for columns to be none
            cols = cls._cast_columns(cols)
        return cls(K=K, mix=mix, means=means, cov_tril=cov_tril, full=full_covariance, columns=cols)

    @property
    def mix(self) -> np.ndarray:
        '''Get the mixture coefficients.'''
        return self._mix

    @property
    def means(self) -> np.ndarray:
        '''Get the means of the Gaussian components.'''
        return self._means

    @property
    def nfeatures(self) -> int:
        '''Get the number of features observed in the dataset.'''
        return self._means.shape[1]

    @property
    def cov_tril(self) -> np.ndarray:
        '''Get the lower triangular matrix of the covariance matrices.'''
        return self._cov_tril

    @property
    def full_covariance(self) -> bool:
        '''Get whether the covariance matrices are full.'''
        return self._full_covariance

    @property
    def K(self) -> int:  # pylint: disable=invalid-name
        '''Get the number of Gaussian components.'''
        return self._K

    def copy(self) -> 'AutonGaussianMixtureModelsNeighbor':
        '''Copy the state of the neighbor.'''
        return self.__class__(K=self._K,
                              mix=self._mix.copy(),
                              means=self._means.copy(),
                              cov_tril=self._cov_tril.copy(),
                              full=self._full_covariance,
                              columns=self.columns)

    @classmethod
    def unflatten(cls,
                  x: np.ndarray,
                  K: int,  # pylint: disable=invalid-name
                  d: int,
                  full: bool) -> 'AutonGaussianMixtureModelsNeighbor':
        '''Unflatten the state of the neighbor.'''
        if not full:
            return cls(K=K,
                       mix=x[:K],
                       means=x[K:(K + K * d)].reshape((K, d)),
                       cov_tril=x[(K + K * d):].reshape((K, d)),
                       full=full)
        cov = x[(K + K * d):]
        cov_tril = np.zeros((K, d, d))
        s = (d * (d + 1)) // 2
        idx = np.tril_indices(d)
        for k in range(K):
            cov_tril[k][idx] = cov[(k * s):((k + 1) * s)]
        return cls(
            K=K,
            full=full, mix=x[:K],
            means=x[K:(K + K * d)].reshape((K, d)),
            cov_tril=cov_tril)

    def flatten(self) -> np.ndarray:
        '''Flatten the state of the neighbor.'''
        if self._cov_tril.ndim == 2:
            return np.concatenate((self._mix, self._means.flatten(), self._cov_tril.flatten()))
        K, d = self._means.shape  # pylint: disable=invalid-name
        idx = np.tril_indices(d)
        return np.concatenate([self._mix, self._means.flatten()]
                              + [self._cov_tril[k][idx] for k in range(K)])

    def jac_mix(self) -> np.ndarray:
        '''Jacobian of the mixture coefficients.'''
        return np.eye(self.mix.size) / self.mix.sum() - self.mix[None, :] / self.mix.sum() ** 2

    def divide_mix_by_sum(self) -> None:
        '''Divide the mixture coefficients by their sum.'''
        self._mix = self._mix / self._mix.sum()

    def log_likelihoods(self, x_: np.ndarray) -> np.ndarray:
        '''
        compute the log likelihood of observations X under each component of a GMM
        '''
        _, d = x_.shape
        _k = self._means.shape[0]
        x_meaned = x_[:, None, :] - self._means[None, :, :]  # (n,.,d),(.,K,d)->(n,K,d)
        logdets = self.logdeterminants()
        if self._cov_tril.ndim == 2:  # diagonal covariances
            logprob = (-0.5 * np.einsum('kd,nkd->nk', 1.0 / self.cov_tril, x_meaned ** 2)
                       - 0.5 * logdets[None, :] - d * 0.5 * np.log(2 * np.pi))
        else:
            z = np.stack([solve_triangular(self._cov_tril[k], x_meaned[:, k, :].T, lower=True)
                          for k in range(_k)],
                         2)  # shape (d,n,K)
            logprob = -0.5 * (z ** 2).sum(0) - 0.5 * logdets[None, :] - d * 0.5 * np.log(2 * np.pi)
        return logprob  # shape=(n, K)

    def log_pdf(self, x_: np.ndarray) -> np.ndarray:
        '''
        compute the log likelihood of observations X under a GMM
        '''
        logprob = self.log_likelihoods(x_)  # shape=(n,K)
        return np.apply_along_axis(logsumexp, 1, logprob + np.log(self._mix)[None, :])

    def grad_lp(self, x, wrt_cov=None) -> Tuple[np.float64, 'AutonGaussianMixtureModelsNeighbor']:
        '''compute the log likelihood of observations X under a GMM and gradients

        compute grad of cov_tril with respect to wrt_cov (i.e. sigma = L@L.T+M@M.T, and want
        gradient with respect to L)
        only needed for full _trilcovariances
        '''
        _, d = x.shape
        K = self._means.shape[0]  # pylint: disable=invalid-name
        x_mean_ed = x[:, None, :] - self._means[None, :, :]  # (n,.,d),(.,K,d)->(n,K,d)
        logdets = self.logdeterminants()
        if self._cov_tril.ndim == 2:  # diagonal covariances
            logprob = (-0.5 * np.einsum('kd,nkd->nk', 1.0 / self._cov_tril, x_mean_ed ** 2)
                       - 0.5 * logdets[None, :] - d * 0.5 * np.log(2 * np.pi))  # shape = (n, K)
            grad_means = x_mean_ed / self.cov_tril[None, :, :]  # shape = (n, K, d)
            grad_cov_tril = (0.5 * (x_mean_ed**2 - self.cov_tril[None, :, :])
                             / self.cov_tril[None, :, :]**2)  # shape = (n, K, d)
        else:
            z = np.stack([solve_triangular(self._cov_tril[k], x_mean_ed[:, k, :].T, lower=True)
                          for k in range(K)],
                         2)  # shape (d, n, K)
            logprob = (-0.5 * (z**2).sum(0)
                       - 0.5 * logdets[None, :]
                       - d * 0.5 * np.log(2 * np.pi))  # shape (n,k)
            grad_means = np.stack([solve_triangular(self._cov_tril[k].T, z[:, :, k], lower=False)
                                   for k in range(K)], 1).T  # shape (n, K, d)
            grad_cov_tril = np.stack(
                [np.einsum('ni,nj->nij', z[:, :, k].T, z[:, :, k].T) - np.eye(d)[None, :, :]
                 for k in range(K)],
                1)  # shape (n,K,d,d)
        #
        weights = logprob + np.log(self._mix)[None, :]  # shape = (n,k)
        mx = weights.max(1)
        weights = np.exp(weights - mx[:, None])
        wtot = weights.sum(1)
        wtotinv = 1.0 / wtot
        #
        lp_ = np.log(wtot).sum() + mx.sum()  # row-wise logsumexp -> scalar
        #
        grad_mix = np.einsum('nk,k,n->k', weights, 1.0 / self.mix, wtotinv)  # shape = (k,)
        #
        grad_means = np.einsum('nk,nki,n->ki', weights, grad_means, wtotinv)  # shape = (k,d)
        #
        if self._cov_tril.ndim == 2:  # diagonal covariances
            grad_cov_tril = np.einsum('nk,nki,n->ki', weights, grad_cov_tril, wtotinv)
        else:
            grad_cov_tril = np.einsum('nk,nkij,n->kij', weights, grad_cov_tril, wtotinv)
            if wrt_cov is not None:
                grad_cov_tril = np.stack([
                    solve_triangular(
                        self._cov_tril[k].T,
                        grad_cov_tril[k]
                        @ solve_triangular(self._cov_tril[k], wrt_cov[k], lower=True),
                        lower=False) for k in range(K)],
                    0)
            else:
                grad_cov_tril = np.stack([
                    solve_triangular(self._cov_tril[k].T, grad_cov_tril[k], lower=False)
                    for k in range(K)],
                    0)
            ri, ci = np.triu_indices(d, k=1)
            for k in range(K):
                grad_cov_tril[k, ri, ci] = 0
        #
        return lp_, self.__class__(
            K=self._K,
            mix=grad_mix,
            means=grad_means,
            cov_tril=grad_cov_tril,
            full=self._full_covariance)

    def norm2(self) -> np.float64:
        '''Compute the norm of the state.'''
        return self.inner_product(self)

    def grad_norm2(self) -> Tuple[np.float64, 'AutonGaussianMixtureModelsNeighbor']:
        '''Compute the gradient of the norm of the state.'''
        # TODO(Piggy): the AutonGaussianMixtureModelsNeighbor this returns represents a gradient
        #       rather than a neighbor's state.
        #       Perhaps use a type alias to make that explicit?
        norm2_, grad_state = self.grad_inner_product(self)
        return norm2_, self.__class__(
            K=self._K,
            mix=2 * grad_state.mix,
            means=2 * grad_state.means,
            cov_tril=2 * grad_state.cov_tril,
            full=self._full_covariance)

    def grad_inner_product(self, other: 'AutonGaussianMixtureModelsNeighbor') -> Tuple[
            np.float64, 'AutonGaussianMixtureModelsNeighbor']:
        '''Compute the inner product of two Gaussian Mixture Models.'''
        K, d = self.means.shape  # pylint: disable=invalid-name
        inner_product_ = np.float64(0.0)
        grad_mix = np.zeros(K)
        grad_means = np.zeros((K, d))
        grad_cov_tril = np.zeros(self._cov_tril.shape)
        cov = self._cov_tril
        for j in range(K):
            if cov.ndim == 2:  # diagonal covariances
                cov = self._cov_tril + other.cov_tril[j][None, :]
            else:
                cov = np.stack([
                    np.linalg.cholesky(self._cov_tril[i] @ self._cov_tril[i].T
                                       + other.cov_tril[j] @ other.cov_tril[j].T)
                    for i in range(K)],
                    0)
            lp_state = self.__class__(
                K=self._K,
                mix=self.mix,
                means=self.means,
                cov_tril=cov,
                full=self._full_covariance)
            lp_, grad_state = lp_state.grad_lp(other.means[j][None, :], self._cov_tril)
            p = np.exp(lp_)
            inner_product_ += other.mix[j] * p
            grad_mix += other.mix[j] * p * grad_state.mix
            grad_means += other.mix[j] * p * grad_state.means
            grad_cov_tril += other.mix[j] * p * grad_state.cov_tril
        return inner_product_, self.__class__(
            K=self._K,
            mix=grad_mix,
            means=grad_means,
            cov_tril=grad_cov_tril,
            full=self._full_covariance)

    def norm2_diff(self, other: 'AutonGaussianMixtureModelsNeighbor') -> np.float64:
        '''Compute numerical representation of distance between models
        (||f-g||^2)
        '''
        retval = self.norm2() + other.norm2() - 2.0 * self.inner_product(other)
        retval = round(retval, ndigits=5)  # Account for floating point error.
        assert retval >= 0.0, (
            'BUG: norm2_diff should never be negative.'
        )
        return retval

    def distance(self, other: 'AutonGaussianMixtureModelsNeighbor') -> float:
        return float(self.norm2_diff(other))

    def grad_norm2_diff(self, other: 'AutonGaussianMixtureModelsNeighbor'
                        ) -> Tuple[np.float64, 'AutonGaussianMixtureModelsNeighbor']:
        '''
        compute grad ||f-g||^2, w.r.t mixture f
        '''
        n2_f, grad_state_f = self.grad_norm2()
        n2_g = other.norm2()
        ip, grad_state_ip = self.grad_inner_product(other)
        n2_diff = n2_f + n2_g - 2.0 * ip
        grad_mix = grad_state_f.mix - 2.0 * grad_state_ip.mix
        grad_means = grad_state_f.means - 2.0 * grad_state_ip.means
        grad_cov_tril = grad_state_f.cov_tril - 2.0 * grad_state_ip.cov_tril
        return n2_diff, self.__class__(
            K=self._K,
            mix=grad_mix,
            means=grad_means,
            cov_tril=grad_cov_tril,
            full=self._full_covariance)

    def logdeterminants(self) -> np.ndarray:
        '''Compute the log determinants of the covariance matrices.'''
        if self._cov_tril.ndim == 2:  # diagonal covariances
            return np.log(self._cov_tril).sum(1)
        K = self._cov_tril.shape[0]  # pylint: disable=invalid-name
        return np.array([self._logdet2d(self._cov_tril[k]) for k in range(K)])

    def _logdet2d(self, cov_tril) -> np.float64:
        return 2 * np.log(np.diag(cov_tril)).sum()

    def inner_product(self, other: 'AutonGaussianMixtureModelsNeighbor') -> np.float64:
        '''Compute the inner product of two Gaussian Mixture Models.'''
        K, d = self._means.shape  # pylint: disable=invalid-name
        inner_product_ = np.float64(0.0)
        for i in range(K):
            for j in range(K):
                if self._cov_tril.ndim == 2:  # diagonal covariances
                    cov = self._cov_tril[i] + other.cov_tril[j]
                    z = (self._means[i] - other.means[j]) / np.sqrt(cov)
                    logdet = np.log(cov).sum()
                else:
                    L = (  # pylint: disable=invalid-name
                        np.linalg.cholesky(self._cov_tril[i] @ self._cov_tril[i].T
                                           + other.cov_tril[j] @ other.cov_tril[j].T)
                    )
                    z = solve_triangular(L, other.means[j] - self._means[i], lower=True)
                    logdet = self._logdet2d(L)
                inner_product_ += (self.mix[i] * other.mix[j]
                                   * np.exp(
                                       -0.5 * z @ z - 0.5 * logdet - d * 0.5 * np.log(2 * np.pi)))
        return inner_product_


class AutonGaussianMixtureModelsInstance(DistributedAlgorithmInstance):
    '''Implementation of Gaussian Mixture Models for distributed data.'''
    _neighbor_constructor = AutonGaussianMixtureModelsNeighbor

    _predict_state: Optional[AutonGaussianMixtureModelsNeighbor] = None

    _K: int  # pylint: disable=invalid-name
    _full_covariance: bool
    _random_state: np.random.RandomState  # pylint: disable=no-member,line-too-long

    def __init__(self,
                 parent: Algorithm,
                 distributed: DistributedConfig,
                 **kwargs):
        hyperparams = parent.hyperparams(**kwargs)
        _K = hyperparams.pop('K')  # pylint: disable=invalid-name
        _full_covariance = hyperparams.pop('full_covariance')
        _random_seed = hyperparams.pop('random_seed', None)
        self._random_state = np.random.RandomState(_random_seed)  # pylint: disable=no-member,line-too-long
        if _K is None:
            raise HyperparamError('K must be specified for Gaussian Mixture Models.')

        if not isinstance(distributed, DistributedConfig):
            raise HyperparamError(
                f'{self.__class__.__name__} distributed must be a DistributedConfig, instead found '
                f'{distributed} of type {type(distributed)}')

        self._K = _K  # pylint: disable=invalid-name
        self._full_covariance = _full_covariance
        self._random_seed = _random_seed
        self._last_dataset = None
        super().__init__(parent, distributed=distributed, **hyperparams)

    @property
    def _neighbor_models_iter(self) -> Iterator[AutonGaussianMixtureModelsNeighbor]:
        for v in super()._neighbor_models_iter:
            assert isinstance(v, AutonGaussianMixtureModelsNeighbor), (
                'BUG: expected neighbor_models to contain AutonGaussianMixtureModelsNeighbor, '
                f'instead found {v} of type {type(v)}'
            )
            yield v

    @property  # type: ignore[override]
    def _my_state(self) -> Optional[AutonGaussianMixtureModelsNeighbor]:
        retval = DistributedAlgorithmInstance._my_state.fget(self)  # type: ignore[attr-defined] # pylint: disable=assignment-from-no-return, line-too-long
        assert retval is None or isinstance(retval, AutonGaussianMixtureModelsNeighbor), (
            'BUG: expected _my_state to be None or an AutonGaussianMixtureModelsNeighbor.')
        return retval

    @_my_state.setter
    def _my_state(self, value: Optional[AutonGaussianMixtureModelsNeighbor]) -> None:
        assert value is None or isinstance(value, AutonGaussianMixtureModelsNeighbor), (
            'BUG: expected value to be None or an AutonGaussianMixtureModelsNeighbor.')
        DistributedAlgorithmInstance._my_state.fset(self, value)  # type: ignore[attr-defined]

    def _decode(self, serialized_model: bytes) -> NeighborState:
        '''Decode a message from distributed neighbors. '''
        return AutonGaussianMixtureModelsNeighbor.decode(serialized_model)

    def _fit(self, dataset: Optional[Dataset], **kwargs) -> None:
        '''Fit the model to the data. This is the actual implementation of fit.'''
        # TODO(Piggy): This function is too long. Refactor it.
        # pylint: disable=too-many-locals, too-many-statements
        _omega = self._omega
        if dataset is None:
            if not self._neighbor_metadata:
                # No data or neighbor models; cannot fit
                return
            logger.debug('No data provided to fit.')
            # No data so we only synthesize neighbor models
            # Set omega to 1 (no self-regularization)
            _omega = np.float64(1.0)
            # Create blank data with 0 rows and a number of columns matching neighbors
            neighbor_models = self._neighbor_models_iter
            ncol: int = next(neighbor_models).nfeatures
            if any(model.nfeatures != ncol for model in neighbor_models):
                raise NotImplementedError(
                    'Neighbor models disagree about number of columns '
                    'and reconciling cols is not implemented.')
            data = np.zeros(shape=(0, ncol))
        else:
            data = dataset.dataframe.to_numpy()
        d = data.shape[1]
        K = self._K  # pylint: disable=invalid-name
        # initial setup
        last_fit = None
        if self._my_state is None:  # first fit
            if dataset is None:
                # Just pick the state of some neighbor.
                self._my_state = next(self._neighbor_models_iter).copy()
            else:
                means = (self._random_state.rand(K, d)
                         * (data.max(0) - data.min(0) + 1e-6)[None, :]
                         + data.min(0) if data.shape[0] > 0 else self._random_state.rand(K, d)
                         - 0.5)
                mix = np.ones(K) / K
                if self._full_covariance:
                    cov_tril = np.zeros((K, d, d)) + np.eye(d)[None, :, :]
                else:
                    cov_tril = np.ones((K, d))
                self._my_state = AutonGaussianMixtureModelsNeighbor(
                    K=K, mix=mix, means=means, cov_tril=cov_tril, full=self._full_covariance,
                    columns=self._columns)
        else:
            last_fit = self._my_state.copy()
        x0 = self._my_state.flatten()

        def fun(x: np.ndarray) -> Tuple[np.float64, np.ndarray]:
            '''Objective function for the fit procedure.'''
            state = AutonGaussianMixtureModelsNeighbor.unflatten(
                x=x, K=K, d=d, full=self._full_covariance)
            jac_mix = state.jac_mix()
            state.divide_mix_by_sum()
            # local data
            lp_, grad_state = state.grad_lp(data)
            jac_grad_state = state.__class__(
                mix=jac_mix @ grad_state.mix,
                means=grad_state.means,
                cov_tril=grad_state.cov_tril,
                full=self._full_covariance)
            grad = jac_grad_state.flatten()

            if last_fit is None:
                return -lp_, -grad
            assert self._my_state is not None, (
                'BUG: model state must be set in fit.'
            )
            state = self._my_state
            # regularization
            reg = np.float64(0.0)
            reg_grad = np.zeros(x.size)
            # self regularization
            nrm2, grad_state = state.grad_norm2_diff(last_fit)
            reg += (1.0 - _omega) / _omega * nrm2
            jac_grad_state = state.__class__(
                mix=jac_mix @ grad_state.mix,
                means=grad_state.means,
                cov_tril=grad_state.cov_tril,
                full=self._full_covariance)
            reg_grad += (1.0 - _omega) / _omega * jac_grad_state.flatten()
            # neighbor regularization
            num_neighbors = len(self._neighbor_metadata)
            for model in self._neighbor_models_iter:
                nrm2, grad_state = state.grad_norm2_diff(model)
                reg += 1.0 / num_neighbors * nrm2
                jac_grad_state = state.__class__(
                    mix=jac_mix @ grad_state.mix,
                    means=grad_state.means,
                    cov_tril=grad_state.cov_tril,
                    full=self._full_covariance)
                reg_grad += (1.0 - _omega) / _omega * jac_grad_state.flatten()
            #
            return np.float64(-lp_ + self._lambda * reg), -grad + self._lambda * reg_grad

        lb = (-np.inf) * np.ones(x0.size)
        ub = (np.inf) * np.ones(x0.size)
        lb[:K] = 1e-6
        ub[:K] = 1.0
        if not self._full_covariance:
            lb[(K + K * d):] = 1e-4
        else:
            ri, ci = np.tril_indices(d)
            # constrain diagonal of cov_tril to be positive
            lb[(K + K * d) + np.where(np.tile(ri == ci, K))[0]] = 1e-2
        bnds = Bounds(lb=lb, ub=ub, keep_feasible=True)
        soln = minimize(fun, x0, method='L-BFGS-B', jac=True, bounds=bnds)
        new_state = self._my_state.__class__.unflatten(
            soln.x, K, d, self._full_covariance)
        new_state.columns = self._columns
        self._my_state = new_state
        self._my_state.divide_mix_by_sum()

    def _predict(self, dataset: Optional[Dataset], **kwargs) -> Optional[Dataset]:
        '''Predict the labels of the data. This is the actual implementation of predict.'''
        if dataset is None:
            return None
        return self.log_pdf(dataset)

    def log_pdf(self, dataset: Dataset) -> Dataset:
        '''Compute the log likelihood of the data under the model.'''
        if self._predict_state is None:
            raise ValueError('Model must be trained before computing log likelihood.')
        data = dataset.dataframe.to_numpy()
        retval = dataset.output()
        retval.dataframe = pd.DataFrame(data=self._predict_state.log_pdf(data))
        return retval

    def mean_log_pdf(self, dataset: Dataset) -> np.float64:
        '''Compute the log likelihood of the data under the model.'''
        if self._predict_state is None:
            raise ValueError('Model must be trained before computing log likelihood.')
        data = dataset.dataframe.to_numpy()
        return self._predict_state.log_pdf(data).mean()

    def score(self, dataset: Dataset) -> np.float64:
        '''Compute the log likelihood of the data under the model.

        Name provided for compatability with sklearn.
        '''
        return self.mean_log_pdf(dataset)

    def sample(self,
               n: int,
               random_state: Optional[np.random.RandomState] = None) -> Dataset:  # pylint: disable=no-member,line-too-long
        '''Sample from the model.

        Args:
            n: The number of samples to generate.
        Returns:
            A sample of size n from the distributed computed in fit().
            A numpy array of shape (n, d) where d is the dimension of the data.
        '''
        # TODO(Piggy): lock sample()
        if not self.trained:
            raise ValueError('Model must be trained before sampling.')
        state = self._my_state
        assert state is not None, (
            'BUG: model state must be set in fit.'
        )

        random_state = random_state or np.random.RandomState(1)  # pylint: disable=no-member,line-too-long
        assert sum(state.mix) == 1
        choices = random_state.choice(len(state.mix), size=n, p=state.mix)
        if self._full_covariance:
            data = np.array(
                [random_state.multivariate_normal(state.means[k],
                                                  state.cov_tril[k] @ state.cov_tril[k].T)
                 for k in choices])
        else:
            data = np.array(
                [random_state.multivariate_normal(state.means[k], np.diag(state.cov_tril[k]))
                 for k in choices])

        metadata = Metadata()
        return Dataset(
            metadata=metadata,
            dataframe=pd.DataFrame(data)
        )

    def norm2_diff(self, other: 'AutonGaussianMixtureModelsInstance') -> np.float64:
        '''Compute the squared norm of the difference between two models.'''
        if self._predict_state is None or other.my_state is None:
            raise ValueError('Model must be trained before computing norm2_diff.')
        return self._predict_state.norm2_diff(other.my_state)

    @property
    def my_state(self) -> Optional[AutonGaussianMixtureModelsNeighbor]:
        '''Get the state of the model.

        Needed for type reconciliation.
        '''
        retval = super().my_state
        assert isinstance(retval, AutonGaussianMixtureModelsNeighbor)
        return retval

    @property
    def my_state_copy(self) -> Optional[AutonGaussianMixtureModelsNeighbor]:
        '''Safely get a copy the state of the model.

        Needed for type reconciliation.
        '''
        retval = super().my_state_copy
        assert isinstance(retval, AutonGaussianMixtureModelsNeighbor)
        return retval


class AutonGaussianMixtureModels(Algorithm):
    '''Implementation of Gaussian Mixture Models for distributed data.'''
    _name = 'auton_gaussian_mixture_models'
    _instance_constructor = AutonGaussianMixtureModelsInstance
    _tags: Dict[str, List[str]] = {
        'task': [TaskType.DENSITY_ESTIMATION.name],
        'data_type': [DataType.TABULAR.name],
        'source': ['autonlab'],
        'distributed': ['true'],
        'supports_random_seed': ['true'],
    }
    _default_hyperparams = {
        'K': None,
        'full_covariance': True,
        'Lambda': 10.0,
        'omega': 2.0 / 3.0,
        'random_seed': Defaults.SEED,
        'tol': 0.000000001,
        'maxiter': None,
    }


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = AutonGaussianMixtureModels(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''tests for dataset_config.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,duplicate-code


import pytest

from ..wrangler.dataset import Column, RoleName
from ..problem_def.problem_def import ProblemDefinition
from .dataset_config import DatasetConfig
from .impl.config_component import ValidationErrors


ROLE_WITH_NO_COL = {
    'config': 'ignore',
    'column_roles': {
        'target': {}
    }
}


def test_role_with_no_col():
    clause = ROLE_WITH_NO_COL.copy()
    dut = DatasetConfig(clause=clause)

    with pytest.raises(ValidationErrors, match=r'[Nn]ame'):
        dut.validate()


INVALID_KEY = {
    'config': 'ignore',
    'hamster': {
        'gerbil': 'gerbil',
    }
}


def test_dataset_invalid_key() -> None:
    clause = INVALID_KEY.copy()
    dut = DatasetConfig(clause=clause)

    with pytest.raises(ValidationErrors, match='hamster'):
        dut.validate()


POS_LABEL_DATASET = {
    'column_roles': {
        'target': {
            'name': 'a',
            'pos_label': 'good'
        }
    }
}


def test_pos_label_sunny_day() -> None:
    dut = DatasetConfig(POS_LABEL_DATASET)
    assert 'good' == dut.metadata.pos_labels[RoleName.TARGET]


def test_pos_label_missing() -> None:
    clause = {
        'column_roles': {
            'target': {
                'name': 'class'
            }
        }
    }
    dut = DatasetConfig(clause=clause)
    assert dut.metadata.pos_labels[RoleName.TARGET] is None


BINARY_CLASSIFICATION_WITH_POS_LABEL = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {
            "target": {
                "name": "class",
                "pos_label": "good"
            }
        }
    },
    "problem_type": {
        "task": "binary_classification"
    }
}
'''


def test_metadata_binary_classification_with_pos_label() -> None:
    dut = ProblemDefinition(BINARY_CLASSIFICATION_WITH_POS_LABEL)
    dataset_conf = dut.get_conf(dut.Keys.DATASET)
    assert isinstance(dataset_conf, DatasetConfig)
    metadata = dataset_conf.metadata
    assert metadata.roles[RoleName.TARGET] == [Column('class')]
    assert metadata.pos_labels[RoleName.TARGET] == 'good'



'''Base class for elements of the Problem Definition file'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
import copy
from typing import Any, Dict, List, Optional, Set, Type, Union

from aenum import Enum as AEnum, extend_enum  # type: ignore[import-untyped]


class ConfigError(BaseException):
    '''All errors thrown by problem definition modules'''


class ConfigDirectoryError(ConfigError, NotADirectoryError):
    '''A path given in the problem doc was not a directory, but should be.'''


class MissingKeyError(ConfigError, KeyError):
    '''We failed to find a given key in the problem doc.'''


class InvalidKeyError(ConfigError, KeyError):
    '''We found an invalid key in the problem doc.'''


class ProblemDefTypeError(ConfigError, TypeError):
    '''A value in the problem definition was the wrong type.'''


class InvalidValueError(ConfigError, ValueError):
    '''A value in the problem definition has the right type but is invalid.'''


class ParsingErrors(ConfigError):
    '''Error(s) thrown when problem definition JSON cannot be parsed'''

    def __init__(self, errors: List[ConfigError]):
        super().__init__(f'At least one parsing error occured: {errors!r}')
        self.errors = errors


class ValidationErrors(ConfigError):
    '''Problem definition JSON can be parsed, but is invalid in some other way.'''

    def __init__(self, errors: List[ConfigError]):
        super().__init__('At least one validation error occured:')
        self.errors = errors

    def __str__(self) -> str:
        retval_list1 = [str(arg) for arg in self.args]
        retval_list2 = [f'\t{repr(error)}' for error in self.errors]
        return '\n'.join(retval_list1 + retval_list2)


class ConfigComponent():
    '''Base class for elements of the Problem Definition file

    This provides _get, _get_with_default, and _exists for
    use by config components.
    '''
    # pylint: disable=too-few-public-methods
    # _parents is a path of ancestor clause names to reach this
    # component, for use in error messages.
    _parents: List[str]
    _clause: Dict[str, Any]
    _sub_component_types: Optional[Dict[str, Type['ConfigComponent']]] = None

    @classmethod
    def register(cls, name: str, config_type: Type['ConfigComponent']) -> None:
        '''Register a new subcomponent.'''
        # Lift the lookup table from the base class to the class receiving the registration.
        if cls._sub_component_types is None:
            cls._sub_component_types = {}
        cls._sub_component_types[name] = config_type  # pylint: disable=unsupported-assignment-operation
        if not hasattr(cls.Keys, name):
            extend_enum(cls.Keys, name, name)

    class Keys(AEnum):
        '''Child classes need to define this class with all their keys.'''

    def __init__(self,
                 clause: Dict[str, Any],
                 parents: Optional[List[str]] = None,
                 **kwargs):
        if parents is None:
            parents = []
        self._parents = parents
        self._clause = clause
        # We need to wait until instantiation time to give all the subcomponents
        # the opportunity to register.
        if self._sub_component_types is not None:
            for name, config_type in self._sub_component_types.items():
                if name in clause:
                    setattr(self,
                            name,
                            config_type(name=name, clause=clause[name], **kwargs))

    def allowed_keys(self) -> Set[str]:
        '''Set of keys that are acceptable for this clause'''
        return {key.value for key in list(self.Keys)}

    # We default to no required keys.
    def required_keys(self) -> Set[str]:
        '''Subset of ALLOWED_KEYS that are required'''
        return set()

    def _add_parent(self,
                    parents: Optional[List[str]],
                    new_parent: str) -> List[str]:
        '''Keep track of where we are in the problem def for error messages.'''
        if parents is None:
            retval = []
        else:
            retval = parents.copy()
        retval.append(new_parent)
        return retval

    def _get(self, *args: Union[int, str, Enum]):
        '''for ex: call _get('metric', 'name') to get metric name'''
        here = self._clause
        so_far = []
        for k in args:
            if isinstance(k, Enum):
                k = k.value
            so_far.append(str(k))
            if isinstance(here, list):
                if not isinstance(k, int):
                    raise ProblemDefTypeError(
                        f'found a list at {".".join(self._parents + so_far[:-1])}, '
                        f'expected an integer at {".".join(self._parents + so_far)}')
                if not -len(here) < k < len(here):
                    raise MissingKeyError(f'Could not find the {k}th element of {here}')
            elif isinstance(here, dict):
                if k not in here:
                    raise MissingKeyError('Could not find '
                                          f'{".".join(so_far)}')
                assert isinstance(k, str), (
                    f'BUG: at _get {".".join(so_far)} we have a non-string key, {k!r} for {here!r}')
            here = here[k]

        return copy.deepcopy(here)

    def _get_with_default(self, *args: Union[int, str, Enum], dflt: Any) -> Any:
        if self._exists(*args):
            return self._get(*args)
        return dflt

    def _exists(self, *args: Union[int, str, Enum]) -> bool:
        '''for ex: call _exists('metric', 'name') to see if metric.name exists'''
        here = self._clause
        so_far = []
        for k in args:
            if isinstance(k, Enum):
                k = k.value
            so_far.append(str(k))
            if isinstance(here, list):
                if not isinstance(k, int):
                    raise ProblemDefTypeError(
                        f'found a list at {".".join(self._parents + so_far[:-1])}, '
                        f'expected an integer at {".".join(self._parents + so_far)}')
                if not -len(here) < k < len(here):
                    raise MissingKeyError(f'Could not find the {k}th element of {here}')
            elif isinstance(here, dict):
                if k not in here:
                    return False
                assert isinstance(k, str), (
                    f'BUG: at _exists {".".join(so_far)} we have a '
                    f'non-string key, {k!r} for {here!r}')
            here = here[k]
        return True

    def validate(self, **kwargs) -> None:
        '''Raise a ConfigError if something about the clause is invalid.

        The base class validate() checks only fotr required and allowed keys.

        For ConfigComponents created by plugins, implementations must not raise an
        error if the clause is an empty dict.
        '''
        _ = kwargs
        errors: List[ConfigError] = []
        if isinstance(self._clause, dict):
            keys = set(self._clause.keys())
            if not self.required_keys().issubset(keys):
                errors.append(MissingKeyError(
                    f'Required keys missing in {".".join(self._parents)} clause: '
                    f'{self.required_keys().difference(keys)}'))

            if not keys.issubset(self.allowed_keys()):
                errors.append(InvalidKeyError(
                    f'Invalid key(s) in {".".join(self._parents)} clause: '
                    f'{keys.difference(self.allowed_keys())}. '
                    f'Valid keys are {self.allowed_keys()}'))

        if len(errors) > 0:
            raise ValidationErrors(errors=errors)


class ConfigComponentStub(ConfigComponent):
    '''Config component for cases where we need a stub.'''
    name = 'stub_config_component'

    def validate(self, **kwargs) -> None:
        pass



'''Stub Data Loader for testing purposes.'''

from typing import Any, Dict, List, Optional

from ..wrangler.dataset import Dataset

from .impl.data_loader import DataLoader
from .impl.data_loader_catalog import DataLoaderCatalog


class DataLoaderStub(DataLoader):
    '''Stub Data Loader'''
    name: str = 'stub'
    tags: Dict[str, List[str]] = {
        'input_format': ['stub_input_format'],
        'loaded_format': ['stub_loaded_format'],
    }

    def _load_train(self) -> Optional[Dataset]:
        return Dataset(metadata=self._metadata)

    def _load_test(self) -> Optional[Dataset]:
        return Dataset(metadata=self._metadata)

    def _load_ground_truth(self) -> Optional[Dataset]:
        return Dataset(metadata=self._metadata)

    def _dataset(self, data: Any, **kwargs) -> Dataset:
        return Dataset(metadata=self._metadata)

    def validate(self, dataset: Optional[Dataset]) -> None:
        pass

    def poll(self, timeout: Optional[float] = 0.0) -> Optional[Dataset]:
        pass

    def build_dataset_from_json(self, json_data: Dict) -> Dataset:
        return Dataset(metadata=self._metadata)


def register(catalog: DataLoaderCatalog):
    '''Register all the objects in this file.'''
    catalog.register(DataLoaderStub)



'''Aggregators for testing purposes'''
from typing import Tuple

import numpy as np


def identity(ranks: np.ndarray,
             reverse: bool = False,
             fail: bool = False) -> Tuple[float, np.ndarray]:
    '''Returns the identity or reversed rank of the first ranking'''
    ranking = ranks[0]
    if fail:
        raise ValueError('intentional failure')
    if reverse:
        return 0., np.flip(ranking)
    return 0., ranking



'''tests for cross_validator_auto.py'''
# pylint: disable=missing-function-docstring,duplicate-code

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..stub_cross_validator import StubCrossValidator
from .cross_validator_auto import CrossValidatorCatalogAuto


def test_sunny_day() -> None:
    dut = CrossValidatorCatalogAuto()
    stub = dut.lookup_by_name('stub_cross_validator')
    assert isinstance(stub, StubCrossValidator)



'''Tests for hyperparam_config.py.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,duplicate-code

import pytest

from ..searcher.param_range import ParamRange
from ..searcher.params import Matcher
from ..wrangler.constants import RangeMethod

from .hyperparam_config import HyperparamConfig, HyperparamTooManyClauses
from .problem_def import remove_comments

TEST_PARAMS = {
    'hyperparams': [
        {
            '_comments': [
                'Apply `max_depth` to all instances of RandomForestClassifier.',
                'Note that algorithm is the name of the algorithm associated with',
                'the step where name is name of the step.'
            ],
            'select': {
                'algorithm': 'sklearn.ensemble.RandomForestClassifier'
            },
            'params': {
                'max_depth': {
                    'fixed': 25
                }
            },
        },
        {
            '_comments': [
                'Apply `strategy` to the algorithm in the "simpleImputer1" step within the',
                'RandomForestClassifier bound pipeline of the tabular_classification template.',
                'This is an unusual special case. You probably want the previous case to set',
                'hyperparams for all instances of the algorithm.'
            ],
            'select': {
                'designator': 'tabular_classifiction::sklearn.ensemble.RandomForestClassifier',
                'name': 'simpleImputer1'
            },
            'params': {
                'strategy': {
                    'fixed': 'median'
                }
            }
        },
        {
            '_comments': [
                'Apply `add_indicator` to all preprocessors of type impute.'
            ],
            'select': {
                'tags': {
                    'preprocess_type': 'impute'
                }
            },
            'params': {
                'add_indicator': {
                    'fixed': 'True'
                }
            }
        },
        {
            '_comments': [
                "Set PRIMITIVE_NAME_1's HYPERPARAM_1 to 1, then 2, then 3",
                "This is how you do grid search. This creates 3 bound pipelines."
            ],
            'select': {
                'algorithm': 'PRIMITIVE_NAME_1'
            },
            'params': {
                'HYPERPARAM_1': {
                    'list': [1, 2, 3]
                }
            }
        },
        {
            '_comments': [
                'Do grid search of values from 0 to 6 in increments of 1 (7 total).',
                'If PRIMITIVE_NAME_1 and PRIMITIVE_NAME_2 are in the same bound',
                'pipeline, we will get 21 bound pipelines.'
            ],
            'select': {
                'algorithm': 'PRIMITIVE_NAME_2'
            },
            'params': {
                'HYPERPARAM_2': {
                    'linear': [0, 1, 6]
                }
            }
        }
    ]
}


def test_hyperparam_sunny_day():
    dut = HyperparamConfig(remove_comments(TEST_PARAMS))
    assert len(dut.overrides) == 5
    overrides = list(dut.overrides)
    assert overrides[0].selector[Matcher.ALGORITHM] == 'sklearn.ensemble.randomforestclassifier'
    assert overrides[0].params['max_depth'] == ParamRange(RangeMethod.FIXED, 25)
    assert overrides[4].params['HYPERPARAM_2'].range == [0, 1, 6]


TOO_MANY_TOP_LEVEL = {
    'hyperparams': [
        {
            'select': {
                'algorithm': 'PRIMITIVE_NAME_1'
            },
            'params': {
                'HYPERPARAM_1': {
                    'list': [1, 2, 3]
                }
            },
            'a_third_thing': {}
        }
    ]
}


def test_hyperparam_extra_clause():
    with pytest.raises(HyperparamTooManyClauses, match='a_third_thing'):
        HyperparamConfig(TOO_MANY_TOP_LEVEL)


TOO_MANY_PARAM_METHODS = {
    'hyperparams': [
        {
            'select': {
                'algorithm': 'PRIMITIVE_NAME_1'
            },
            'params': {
                'HYPERPARAM_1': {
                    'list': [1, 2, 3],
                    'fixed': 25,
                }
            },
        }
    ]
}


def test_hyperparam_extra_method():
    with pytest.raises(HyperparamTooManyClauses, match='fixed'):
        HyperparamConfig(TOO_MANY_PARAM_METHODS)


NO_GRID_PARAMS = {
    'hyperparams': [
        "disable_grid_search",
        {
            'select': {
                'algorithm': 'sklearn.ensemble.RandomForestClassifier'
            },
            'params': {
                'max_depth': {
                    'list': [1, 2, 3],
                    "default": 1,
                }
            },
        },
    ]
}


def test_disable_grid_search() -> None:
    dut = HyperparamConfig(NO_GRID_PARAMS)
    assert dut.disable_grid_search
    assert len(dut.overrides) == 1



'''Executes executable pipelines and returns results'''
import abc
from typing import Dict, Optional, Union

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .cucumber import JarOfCucumbers
from .executor_kind import ExecutorKind
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import (ExecutablePipeline, PipelineResult,
                                                FitError)
from ..wrangler.dataset import Dataset


class TrainedPipelineCollection(Dict[Designator, Union[FitError, JarOfCucumbers]]):
    '''The trained models for each pipeline indexed by pipeline name.'''

    def models(self, name: Designator) -> Union[FitError, JarOfCucumbers]:
        '''The models for a pipeline with the given name.'''
        return self[name]

    def set(self, name: Designator, models: Union[FitError, JarOfCucumbers]) -> None:
        '''Save a trained model at a step name.'''
        self[name] = models


class Executor(metaclass=abc.ABCMeta):
    '''Base class for pipeline executors'''
    _kind: ExecutorKind

    @property
    def kind(self) -> ExecutorKind:
        '''Documents the kind of ExecutablePipeline we need.'''
        return self._kind

    @abc.abstractmethod
    def fit(self,
            dataset: Optional[Dataset],
            pipelines: Dict[Designator, ExecutablePipeline]
            ) -> TrainedPipelineCollection:
        '''Train a list of pipelines on a dataset.'''

    @abc.abstractmethod
    def predict(self,
                dataset: Dataset,
                pipelines: Dict[Designator, ExecutablePipeline]
                ) -> Dict[Designator, PipelineResult]:
        '''Use a list of pipelines to predict from a dataset.'''


class ExecutorStub(Executor):
    '''stub'''
    _kind = ExecutorKind('stub_executor_kind')

    def fit(self,
            dataset: Optional[Dataset],
            pipelines: Dict[Designator, ExecutablePipeline]
            ) -> TrainedPipelineCollection:
        '''Train a list of pipelines on a dataset.'''
        return TrainedPipelineCollection()

    def predict(self,
                dataset: Dataset,
                pipelines: Dict[Designator, ExecutablePipeline]
                ) -> Dict[Designator, PipelineResult]:
        '''Use a list of pipelines to predict from a dataset.'''
        return {d: PipelineResult(executable_pipeline=p, prediction=dataset)
                for d, p in pipelines.items()}



'''Test the CommunicatorCatalog'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..stub_communicator import CommunicatorStub
from .communicator_catalog import CommunicatorCatalogStub

# pylint: disable=missing-function-docstring,duplicate-code


def test_register() -> None:
    dut = CommunicatorCatalogStub()
    assert 'stub_communicator' == dut.register(obj=CommunicatorStub)


def test_register_name() -> None:
    dut = CommunicatorCatalogStub()
    stub_communicator = CommunicatorStub
    assert 'somenamedcommunicator' == dut.register(
        obj=stub_communicator, name='SomeNamedCommunicator')


def test_lookup_by_name() -> None:
    dut = CommunicatorCatalogStub()
    some_communicator = CommunicatorStub
    dut.register(obj=some_communicator, name='SomeNamedCommunicator')
    assert dut.lookup_by_name('SomeNamedCommunicator') == some_communicator


def test_lookup_one_by_tag() -> None:
    dut = CommunicatorCatalogStub()
    dut.register(obj=CommunicatorStub, name='NamedStubCommunicator', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 1


def test_lookup_by_tag_fail() -> None:
    dut = CommunicatorCatalogStub()
    dut.register(obj=CommunicatorStub, name='NamedStubCommunicator', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='different_value')) == 0


def test_lookup_multi_by_tag() -> None:
    dut = CommunicatorCatalogStub()
    stub_communicator = CommunicatorStub
    dut.register(obj=stub_communicator, name="Name1", tags={'k1': ['v1']})
    stub_communicator2 = CommunicatorStub
    dut.register(obj=stub_communicator2, name="Name2", tags={'k1': ['v1']})
    stub_communicator3 = CommunicatorStub
    dut.register(obj=stub_communicator3, name="Name3", tags={'k1': ['different_value']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 2



'''Tests for sklearn_loader.py and sklearn_laoders.py'''
from copy import deepcopy

import pandas as pd

from ...config_components.dataset_config import DatasetConfig
from ...problem_def.problem_def import ProblemDefinition
from ...wrangler.wrangler import Wrangler
from ..impl.data_loader_auto import DataLoaderCatalogAuto
from ..impl.dataframe_loader import DataframeLoader
from .sklearn_loader_generator import SklearnLoaderGenerator

# pylint: disable=missing-function-docstring, duplicate-code


MOONY_CLAUSE = {
    'input_format': 'sklearn.datasets.make_moons',
    'params': {
        'n_samples': 5,
        'shuffle': True,
        'noise': None
    }
}


def test_moony_day() -> None:
    dut = SklearnLoaderGenerator(
        name='sklearn.datasets.make_moons',
        tags={
            'input_format': ['sklearn.datasets.make_moons'],
            'loaded_format': ['pandas_dataframe'],
            'supports_random_seed': ['true']
        }
    )

    got = dut.new_class()

    config = DatasetConfig(clause=MOONY_CLAUSE)

    # pylint is wrong about this class not defining the abstract methods.
    inst = got(config=config)  # pylint: disable=abstract-class-instantiated

    assert isinstance(inst, DataframeLoader)

    got_data = inst.load_train()
    want_df = pd.DataFrame({
        '0': [-1.0, 1, 1, 0, 2],
        '1': [0.0, 0, -0.5, 0.5, 0.5],
        'y': [0, 0, 1, 1, 1]
    })
    assert got_data is not None
    pd.testing.assert_frame_equal(got_data.dataframe, want_df)


def test_set_seed() -> None:
    dut = SklearnLoaderGenerator(
        name='sklearn.datasets.make_moons',
        tags={
            'input_format': ['sklearn.datasets.make_moons'],
            'loaded_format': ['pandas_dataframe'],
            'supports_random_seed': ['true']
        }
    )

    got = dut.new_class()

    clause = deepcopy(MOONY_CLAUSE)
    assert isinstance(clause['params'], dict)
    clause['params']['random_seed'] = 1337
    config = DatasetConfig(clause=clause)

    # ibid
    inst = got(config=config)  # pylint: disable=abstract-class-instantiated
    assert isinstance(inst, DataframeLoader)

    got_data = inst.load_train()
    want_df = pd.DataFrame({
        '0': [1.0, 0, -1, 1, 2],
        '1': [-0.5, 0.5, 0, 0, 0.5],
        'y': [1, 1, 0, 0, 1]
    })
    assert got_data is not None
    pd.testing.assert_frame_equal(got_data.dataframe, want_df)


def test_lookup() -> None:
    config = DatasetConfig(clause=MOONY_CLAUSE)
    dut = DataLoaderCatalogAuto().construct_instance(config=config)

    got_data = dut.load_train()
    want_df = pd.DataFrame({
        '0': [-1.0, 1, 1, 0, 2],
        '1': [0.0, 0, -0.5, 0.5, 0.5],
        'y': [0, 0, 1, 1, 1]
    })
    assert got_data is not None
    pd.testing.assert_frame_equal(got_data.dataframe, want_df)
    assert dut.__class__.__name__.startswith('GeneratedSklearnLoader')


DIABETES_CLAUSE = {
    'input_format': 'sklearn.datasets.load_diabetes',
    'params': {
        'scaled': True
    }
}


def test_diabetes() -> None:
    dut = SklearnLoaderGenerator(
        name='sklearn.datasets.load_diabetes',
        tags={
            'input_format': ['sklearn.datasets.load_diabetes'],
            'loaded_format': ['pandas_dataframe'],
            'supports_random_seed': ['false'],
            'uses_return_X_y': ['true'],
            'uses_as_frame': ['true']
        }
    )
    got = dut.new_class()

    clause = deepcopy(DIABETES_CLAUSE)
    config = DatasetConfig(clause=clause)

    # ibid
    inst = got(config=config)  # pylint: disable=abstract-class-instantiated
    assert isinstance(inst, DataframeLoader)

    got_data = inst.load_train()
    assert got_data is not None
    assert got_data.dataframe.shape == (442, 11)
    assert list(got_data.dataframe.columns) == [
        'age', 'sex', 'bmi', 'bp', 's1', 's2', 's3', 's4', 's5', 's6', 'y'
    ]


def test_restrict_features_integration() -> None:
    dut = Wrangler(ProblemDefinition(
        {
            "dataset": {
                "input_format": ["sklearn.datasets.load_breast_cancer"],
                "loaded_format": ["pandas_dataframe"],
                "column_roles": {
                    "target": {
                        "name": "target"
                    }
                },
                "params": {
                    "restrict_features": 2
                }
            },
            "problem_type": {
                "task": "binary_classification"
            }
        }
    ))
    got = dut.load_train_dataset()
    assert got is not None
    got_df = got.dataframe
    assert got_df.shape == (569, 3)
    got_cols = list(got_df.columns)
    assert got_cols == ['mean radius', 'mean texture', 'target']

    assert dut.load_test_dataset() is None
    assert dut.load_ground_truth_dataset() is None


def test_train_test_split_integration() -> None:
    dut = Wrangler(ProblemDefinition(
        {
            "dataset": {
                "input_format": ["sklearn.datasets.load_breast_cancer"],
                "loaded_format": ["pandas_dataframe"],
                "column_roles": {
                    "target": {
                        "name": "target"
                    }
                },
                "params": {
                    "test_size": 2,
                }
            },
            "problem_type": {
                "task": "binary_classification"
            }
        }
    ))
    got_train = dut.load_train_dataset()
    assert got_train is not None
    got_train_df = got_train.dataframe
    assert got_train_df.shape == (567, 31)

    got_test = dut.load_test_dataset()
    assert got_test is not None
    got_test_df = got_test.dataframe
    assert got_test_df.shape == (2, 30)

    got_gt = dut.load_ground_truth_dataset()
    assert got_gt is not None
    got_gt_df = got_gt.ground_truth
    assert got_gt_df.shape == (2, 1)



'''Test the binarizer model.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd

from ..wrangler.dataset import Column, Dataset, Metadata, RoleName

from .binarizer import BinarizerModel
# pylint: disable=missing-function-docstring,duplicate-code


def test_sunny_day() -> None:
    model = BinarizerModel()

    metadata = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        pos_labels={RoleName.TARGET: 'good'})

    dataset = Dataset(
        metadata=metadata,
        dataframe=pd.DataFrame({
            'target': [b'good', b'bad', b'good'],
            'other': [b'good', b'indifferent', b'middlin'],
        })
    )

    dut = model.instantiate()
    got = dut.predict(dataset)

    want = Dataset(
        metadata=metadata,
        dataframe=pd.DataFrame({
            'target': [True, False, True],
            'other': [b'good', b'indifferent', b'middlin'],
        })
    )

    pd.testing.assert_frame_equal(got.get_dataframe(), want.get_dataframe())


def test_set_pos_from_constructor() -> None:
    model = BinarizerModel(target='good')

    metadata = Metadata(roles={RoleName.TARGET: [Column('target')]})

    dataset = Dataset(
        metadata=metadata,
        dataframe=pd.DataFrame({
            'target': [b'good', b'bad', b'good'],
            'other': [b'good', b'indifferent', b'middlin'],
        })
    )

    dut = model.instantiate()
    got = dut.predict(dataset)

    want = Dataset(
        metadata=metadata,
        dataframe=pd.DataFrame({
            'target': [True, False, True],
            'other': [b'good', b'indifferent', b'middlin'],
        })
    )

    pd.testing.assert_frame_equal(got.get_dataframe(), want.get_dataframe())



'''Class representing problem type, which is a combination of data type and task.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum, auto
from typing import Any, Dict, List, Optional, Set

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..config_components.impl.config_component import (
    ConfigComponent, ConfigError, ValidationErrors,
    InvalidValueError, ProblemDefTypeError)
from ..wrangler.constants import ProblemDefKeys, Defaults


class DataType(Enum):
    '''Possible input data types'''
    TABULAR = auto()
    IMAGE = auto()
    TIMESERIES = auto()
    TEXT = auto()

    @staticmethod
    def validate(name: str) -> List[str]:
        '''Return an error message if data type is not in the list'''
        errors: List[str] = []
        names = [member.name for member in DataType]
        if name.upper() not in names:
            errors.append(f'"{name}" not among valid data types: f{names}')
        return errors


class TaskType(Enum):
    '''Possible ML tasks'''
    BINARY_CLASSIFICATION = auto()
    MULTICLASS_CLASSIFICATION = auto()
    REGRESSION = auto()
    FORECASTING = auto()
    DENSITY_ESTIMATION = auto()
    OTHER = auto()
    TEST_TASK = auto()  # Only used in testing

    @staticmethod
    def validate(name: str) -> List[str]:
        '''Return an error message if task type is not in the list'''

        # TODO(Merritt/Piggy): figure out how to deal with this without using forecasting
        errors: List[str] = []
        names = [member.name for member in TaskType]
        if name.upper() not in names:
            errors.append(f'"{name}" not among valid task types: f{names}')
        return errors


class Task(ConfigComponent):
    '''Parsed form of the problem_type.'''

    class Keys(AEnum):
        '''Top-level keys in Task component'''
        DATA_TYPE = 'data_type'
        TASK_TYPE = 'task'

    def required_keys(self) -> Set[str]:
        '''Subset of ALLOWED_KEYS that are required'''
        return {self.Keys.TASK_TYPE.value}  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long

    def __init__(self, clause: Dict[str, Any], parents: Optional[List[str]] = None):
        parents = self._add_parent(parents, ProblemDefKeys.TASK.value)
        super().__init__(clause=clause, parents=parents)

    @property
    def data_type(self) -> DataType:
        '''Data type'''
        return DataType[self._get_with_default(
            self.Keys.DATA_TYPE,
            dflt=Defaults.DATA_TYPE).upper()]

    @property
    def task_type(self) -> TaskType:
        '''ML task'''
        field = self._get(self.Keys.TASK_TYPE)
        try:
            return TaskType[field.upper()]
        except KeyError as err:
            raise InvalidValueError(
                f'task type {field} is not recognized.') from err

    def __str__(self) -> str:
        data_type_name = "None"
        if self.data_type is not None:
            data_type_name = self.data_type.name
        task_name = "None"
        if self.task_type is not None:
            task_name = self.task_type.name
        return (f'{{Data type: {data_type_name}'
                f', Task: {task_name} }}')

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, Task):
            if __value.data_type == self.data_type and __value.task_type == self.task_type:
                return True
        return False

    def validate(self, **kwargs) -> None:
        # test that problem_type maps to a dictionary containing exactly data_type and task_type
        super().validate(**kwargs)
        errors: List[ConfigError] = []

        if isinstance(self._clause, dict):

            if self.Keys.DATA_TYPE.value in self._clause:  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
                errmessages = DataType.validate(self._clause[self.Keys.DATA_TYPE.value])  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
                errors = errors + [InvalidValueError(err) for err in errmessages]

            if self.Keys.TASK_TYPE.value in self._clause:  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
                errmessages = TaskType.validate(self._clause[self.Keys.TASK_TYPE.value])  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
                errors = errors + [InvalidValueError(err) for err in errmessages]
        else:
            errors.append(ProblemDefTypeError(
                f'{ProblemDefKeys.TASK.value} must be a dict, '
                f'instead found {type(self._clause)}'))

        if len(errors) > 0:
            raise ValidationErrors(errors=errors)



""" Tests for memory_data_loader.py"""
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from copy import deepcopy
import pytest

import pandas as pd

from ..config_components.dataset_config import DatasetConfig
from ..config_components.impl.config_component import ConfigError
from ..wrangler.dataset import RoleName

from .memory_data_loader import MemoryDataLoader

# pylint: disable=missing-function-docstring,duplicate-code
# Note that we disable flake8 error F841 (# noqa F841)
# because we have a local variable which is used
# but which static analysis can't tell is used.


CLAUSE = {
    'config': 'memory',
    'column_roles': {
        'target': {
            'name': 'a',
            'pos_label': 1
        }
    },
    'params': {
        'train_data': 'train_df'
    }
}
TRAIN_DF = pd.DataFrame({'a': [1, 2], 'b': [3, 4]})


def test_sunny_day():
    train_df = TRAIN_DF  # noqa F841 pylint: disable=unused-variable
    # this needs to be a local variable so the loader picks it up
    config = DatasetConfig(clause=CLAUSE)
    dut = MemoryDataLoader(config=config)
    result = dut.load_train()
    got_train_df = result.dataframe

    assert got_train_df.shape == (2, 2)
    assert RoleName.TARGET in result.metadata.roles.keys()
    assert len(result.metadata.roles[RoleName.TARGET]) == 1
    assert result.metadata.roles[RoleName.TARGET][0].name == 'a'


def test_no_target():
    '''Test that data with no target is acceptable'''
    train_df = TRAIN_DF  # noqa  F841 pylint: disable=unused-variable
    clause_without_target = CLAUSE.copy()
    del clause_without_target['column_roles']
    config = DatasetConfig(clause=clause_without_target)
    dut = MemoryDataLoader(config=config)

    assert dut.load_train().metadata.target is None


def test_load_testdata() -> None:
    train_df = TRAIN_DF  # noqa F841 pylint: disable=unused-variable
    load_test_df = pd.DataFrame(  # noqa F841 pylint: disable=unused-variable
        {'a': [1, 2, 2], 'b': [30, 40, 500]})
    clause = deepcopy(CLAUSE)
    assert isinstance(clause['params'], dict)
    clause['params']['test_data'] = 'load_test_df'
    config = DatasetConfig(clause=clause)
    dut = MemoryDataLoader(config=config)
    result = dut.load_test()

    assert result is not None   # needed to make pylint happy
    got_test_df = result.dataframe

    assert got_test_df.shape == (3, 1)
    assert 'a' not in got_test_df.columns
    assert 'b' in got_test_df.columns


def test_load_nonexistent_testdata() -> None:
    train_df = TRAIN_DF  # noqa F841 pylint: disable=unused-variable
    config = DatasetConfig(clause=CLAUSE)
    dut = MemoryDataLoader(config=config)
    result = dut.load_test()

    assert result is None


MEMORY_CONFIG = {
    "config": "memory",
    "column_roles": {
        "target": {
            "name": "c"
        }
    },
}


def test_missing_dataframe() -> None:
    config = DatasetConfig(clause=MEMORY_CONFIG)
    with pytest.raises(ConfigError, match=r'(memory.*train_data)|(train_data.*memory)/i'):
        MemoryDataLoader(config=config)



'''Decider for tests that always decides to send.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Mapping
from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..neighbor_manager.node_id import NodeID
from ..config_components.distributed_config import DecidersConfig, DeciderConfig

from .impl.decider import Decider
from .impl.decider_catalog import DeciderCatalog


class DeciderStubConfig(DeciderConfig):
    '''Configuration for DeciderStub.'''


class DeciderStub(Decider):
    '''Stub Decider.'''
    name = 'stub_decider'
    tags = {}

    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''
        return True


def register(catalog: DeciderCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(DeciderStub)
    DecidersConfig.register(name=DeciderStub.name, config_type=DeciderStubConfig)



'''Tests for json_instantiator.py'''
import json
from typing import Any
import re

import pytest

from ..algorithms.connect import ConnectorModel
from ..algorithms.impl.fake_algorithm import FakeAlgorithm
from ..executor.cucumber import Cucumber, JarOfCucumbers
from ..generator.bound_pipeline import BoundPipelineStub
from ..generator.designator import Designator, StepDesignator
from ..problem_def.output_config import OutputConfig
from ..wrangler.saver import Saver
from .instantiator import InstantiatorStub
from .json_instantiator import JsonInstantiator

# pylint: disable=missing-class-docstring, missing-function-docstring, redefined-outer-name, duplicate-code


@pytest.fixture(scope='session')
def output_config(tmp_path_factory: pytest.TempPathFactory) -> OutputConfig:
    tmp_path = tmp_path_factory.mktemp('data')
    tmp_dir = tmp_path / 'sub'
    tmp_dir.mkdir()
    config = OutputConfig(clause={
        'path': str(tmp_dir),
        'instantiations': [
            'json'
        ]
    })
    return config


@pytest.fixture(scope='session')
def json_instantiator(tmp_path_factory: pytest.TempPathFactory) -> JsonInstantiator:
    return JsonInstantiator(saver=Saver(output_config=output_config(tmp_path_factory)))


KEYS_TO_TRANSFORM = {'model_filename'}


def strip_serial(top: Any) -> Any:
    '''For a parsed JSON representation of a pipeline, replace all serial
    numbers in model filenames with "NUM" (since we cannot otherwise predict them)'''
    retval = top
    if isinstance(top, dict):
        retval = {}
        for key, value in top.items():
            if key in KEYS_TO_TRANSFORM:
                retval[key] = re.sub('[0-9]+', 'NUM', value)
                continue
            retval[key] = strip_serial(value)
        return retval
    if isinstance(top, list):
        retval = []
        for item in top:
            retval.append(strip_serial(item))
        return retval
    return retval


def test_json_instantiator_sunny_day(output_config: OutputConfig) -> None:
    pipeline = BoundPipelineStub(name='Fake_Bound')
    pipe1 = pipeline.new(name='at')
    pipe1.step(FakeAlgorithm(name='A')).mark_queried()
    pipe1.step(FakeAlgorithm(name='T'))
    pipe2 = pipeline.new(name='dt')
    pipe2.step(FakeAlgorithm(name='D')).set_name('named_step')
    pipe2.step(FakeAlgorithm(name='T'))
    pipeline.parallel(left=pipe1, right=pipe2)
    pipeline.step(ConnectorModel(
        name='ring',
        newkey=['left', 'oldkey'],
        othernewkey=['right', 'oldkey']))
    pipeline.validate()

    # Create a fake JarOfCucumbers for the saver to save.
    cucumbers = JarOfCucumbers({
        StepDesignator('a'): Cucumber(
            impl=FakeAlgorithm().instantiate(),
            filename=StepDesignator('a'),
            pipeline_designator=pipeline.designator,
            hyperparams={}
        )
    })

    # Create a fake executable pipeline, storing that fake jar of cucumbers.
    compiler = InstantiatorStub(cucumbers)
    pipelines = {
        Designator('Fake_Bound'): compiler.instantiate(pipeline=pipeline),
    }

    # Make a saver and save the models in our fake executable pipeline,
    #   writing the pickled fake cucumbers to disk.
    saver = Saver(output_config=output_config)
    model_paths = saver.save_models(pipelines=pipelines)

    # Use the JsonInstantiator to save our fake executable pipeline as json
    #   Uses the paths to the fake saved models provided by the saver
    dut = JsonInstantiator(saver=saver)
    dut.save(pipeline, model_paths=model_paths)

    assert output_config.path is not None
    with open(output_config.path / 'pipelines' / 'fake_bound.json', 'rb') as file_pointer:
        got = json.load(file_pointer)

    assert strip_serial(got) == {
        'version': '1.0',
        'pipeline_designator': 'fake_bound',
        'pipeline_template_name': 'fake_bound',
        'output_dir': str(output_config.path),
        'steps': [
            {
                'type': 'parallel',
                'subpipelines': {
                    'left': [
                        {
                            'algorithm_name': 'a',
                            'model_filename': '@a_NUM@',
                            'type': 'step',
                            'queried': True
                        },
                        {
                            'algorithm_name': 't',
                            'model_filename': '@t_NUM@',
                            'type': 'step'
                        }
                    ],
                    'right': [
                        {
                            'algorithm_name': 'd',
                            'model_filename': '@d_NUM@named_step',
                            'type': 'step',
                            'name': 'named_step'
                        },
                        {
                            'algorithm_name': 't',
                            'model_filename': '@t_NUM@',
                            'type': 'step'
                        }
                    ]
                },
            },
            {
                'type': 'step',
                'algorithm_name': 'ring',
                'hyperparams': {
                    'newkey': ['left', 'oldkey'],
                    'othernewkey': ['right', 'oldkey']
                },
                'model_filename': '@ring_NUM@',
            }
        ],
    }



import numpy as np


def check_theta_phi(theta, phi):
    """This function automatically converts theta to phi or phi to theta as
    list or float depending on the values and value types given as input.
        Parameters
        ----------
        theta: float or list
            Dispersion parameter theta to convert to phi (can be None)
        phi: float or list
            Dispersion parameter phi to convert to theta (can be None)
        Returns
        -------
        tuple
            tuple containing both theta and phi (of list or float type depending on the input type)
    """
    if not ((phi is None) ^ (theta is None)):
        print("Set valid values for phi or theta")
    if phi is None and not isinstance(theta, list):
        phi = theta_to_phi(theta)
    if theta is None and not isinstance(phi, list):
        theta = phi_to_theta(phi)
    if phi is None and isinstance(theta, list):
        phi = [theta_to_phi(t) for t in theta]
    if theta is None and isinstance(phi, list):
        theta = [phi_to_theta(p) for p in phi]
    return np.array(theta), np.array(phi)


def theta_to_phi(theta):
    """ This functions converts theta dispersion parameter into phi
        Parameters
        ----------
        theta: float
            Real dispersion parameter
        Returns
        -------
        float
            phi real dispersion parameter
    """
    return np.exp(-theta)


def phi_to_theta(phi):
    """This functions converts phi dispersion parameter into theta
        Parameters
        ----------
        phi: float
            Real dispersion parameter
        Returns
        -------
        float
            theta real dispersion parameter
    """
    return -np.log(phi)



'''Pipeline loader that loads a single algorithm.'''

from ..generator.bound_pipeline import BoundPipeline
from .impl.pipeline_loader import (
    MissingArgumentsError, PipelineLoader, UnknownArgumentsError)
from .impl.pipeline_loader_catalog import PipelineLoaderCatalog


class Just(PipelineLoader):
    '''Load a pipeline with Just one algorithm.'''
    name = 'just'
    tags = {}

    def _load(self, *args, alg: str = '', **kwargs) -> BoundPipeline:
        '''Load the pipeline.'''

        if alg == '':
            if len(args) > 0:
                alg = args[0]
                args = args[1:]
            else:
                raise MissingArgumentsError(
                    'Missing argument "alg" (algorithm name) '
                    f'to {self.__class__.__name__}'
                )

        if len(args) > 0 or len(kwargs) > 0:
            raise UnknownArgumentsError(
                f'Found unknown arguments to {self.__class__.__name__}:'
                f'*args={args}, **kwargs={kwargs}'
            )

        assert self._algorithm_catalog is not None, f'algorithm_catalog is required for {self.name}'

        algorithm = self._algorithm_catalog.lookup_by_name(alg)

        pipeline = BoundPipeline(name=algorithm.name, tags={})
        pipeline.step(model=algorithm)

        return pipeline


def register(catalog: PipelineLoaderCatalog, **kwargs):
    '''Register all the metrics in this file.'''
    catalog.register(Just(**kwargs))



'''Configuration for the hyperparam clause.

 We use a single data structure, HyperparamConfig, for all hyperparam clauses
 instead of separate structures for each clause so that it is easier to do
 intelligent searches of hyperparameter overrides.
'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
from typing import Dict, Any, List, Union

from ..config_components.impl.config_component import ConfigComponent, ConfigError
from ..searcher.params import Overrides, Override, Selector, ParamRanges, ParamRange
from ..wrangler.constants import (ProblemDefControl,
                                  RangeMethod, Matcher, MATCHERS_TO_LOWERCASE)


class HyperparamError(ConfigError):
    '''Accummulating error for hyperparam parsing.'''

    errors: List[ConfigError]

    def __init__(self, errors: List[ConfigError]):
        super().__init__(f'At least one validation error occured: {errors!r}')
        self.errors = errors


class HyperparamTooManyClauses(ConfigError):
    '''A subclause had unexpected keys.'''


class ControlError(ConfigError):
    '''The control strings for hyperparam config are incorrect.'''


class HyperparamConfig(ConfigComponent):
    '''Configuration for the hyperparam clause.

    We use a single data structure, HyperparamConfig, for all hyperparam clauses
    instead of separate structures for each clause so that it is easier to do
    intelligent searches of hyperparameter overrides.
    '''
    _control: List[str]
    _disable_grid_search = False

    class Constants(Enum):
        '''Keys below the top level.'''
        HYPERPARAMS = 'hyperparams'
        SELECT = 'select'
        PARAMS = 'params'
        DEFAULT = 'default'

    def __init__(self, clause: Dict[str, Any]):
        super().__init__(clause=clause)
        self._overrides = Overrides()

        (overrides, control) = self._split_overrides_control(
            self._get(self.Constants.HYPERPARAMS))
        for override_clause in overrides:
            self._overrides.append(self._parse_override(override_clause))
        self._control = control
        self._parse_control()

        self.validate()

    def _parse_control(self):
        for s in self._control:
            if s == ProblemDefControl.DISABLE_GRID_SEARCH.value:
                self._disable_grid_search = True
            else:
                raise ControlError(f'Invalid control {s}')

    def _split_overrides_control(self, clauses: List[Union[str, Dict[str, Any]]]):
        retval_str = []
        retval_overrides = []
        for val in clauses:
            if isinstance(val, str):
                retval_str.append(val)
            elif isinstance(val, dict):
                retval_overrides.append(val)
            else:
                raise ControlError(f'Unexpected hyperparam type, {val}:{type(val)}')
        return (retval_overrides, retval_str)

    def _parse_override(self, subclause: Dict[str, Dict[str, Any]]) -> Override:
        select_clause = subclause.pop(self.Constants.SELECT.value)
        params_clause = subclause.pop(self.Constants.PARAMS.value)
        if subclause:
            raise HyperparamTooManyClauses(f'Unexpected extra clauses: {subclause!r}')
        select = Selector()
        for matcher, arg in select_clause.items():
            matcher_enum = Matcher[matcher.upper()]
            if matcher_enum in MATCHERS_TO_LOWERCASE:
                arg = arg.lower()
            select[matcher_enum] = arg
        params = ParamRanges()
        for name, hyperparam_param in params_clause.items():
            params[name] = self._parse_param(hyperparam_param)

        return Override(selector=select, params=params)

    def _parse_param(self, param: Dict[str, Any]) -> ParamRange:
        default = param.pop(self.Constants.DEFAULT.value, None)
        if len(param) != 1:
            raise HyperparamTooManyClauses(f'expecting 1 key, got {param!r}')

        for method, range_value in param.items():
            return ParamRange(method=RangeMethod[method.upper()],
                              prange=range_value,
                              default=default)
        assert False, 'BUG: Unreachable code'

    @property
    def disable_grid_search(self) -> bool:
        """Should we disable grid search?

        If true, we skip the Searcher.
        """
        return self._disable_grid_search

    @property
    def overrides(self) -> Overrides:
        '''Return the hyperparam overrides described by the config.'''
        return self._overrides

    def validate(self, **kwargs) -> None:
        '''Make sure the hyperparam object makes sense.'''
        allowed = {x.value for x in ProblemDefControl}
        if not set(self._control).issubset(allowed):
            not_recognized = set(self._control) - allowed
            raise ControlError(
                f'Unrecognized control options: {not_recognized}'
                f'Must be among {allowed}'
            )



'''Test the TSAD aggregators'''

# pylint: disable=duplicate-code

from typing import Optional

from ..generator.bound_pipeline import BoundPipelineStub
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import PipelineResult, PipelineResults
from ..metrics.impl.metric import Metric
from ..ranker.ranker import Ranking, Rankings
from ..wrangler.dataset import Dataset

from .impl.aggregate_ranker import AggregateRanker
from .impl.aggregator_auto import AggregatorCatalogAuto
from .tsad_aggregator import TSADAggregator


class MetricMinVal(Metric):
    '''Test metric returning the minimum of the predictions.'''
    _name = 'min_val_metric'
    _high = True

    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        return min(pred.values())


def test_tsad_simple_ranking() -> None:
    '''Tests the overriding of hyperparameters works for aggregators.'''
    catalog = AggregatorCatalogAuto()
    aggregator = catalog.lookup_by_name('test_aggregators.identity')

    results_1 = Dataset({
        '1': 1,
    })
    results_2 = Dataset({
        '1': 2,
    })
    results_3 = Dataset({
        '1': 3,
    })

    pipeline_results = PipelineResults({
        Designator('pipeline_1'): PipelineResult(prediction=results_1,
                                                 bound_pipeline=BoundPipelineStub('pipeline_1')),
        Designator('pipeline_2'): PipelineResult(prediction=results_2,
                                                 bound_pipeline=BoundPipelineStub('pipeline_2')),
        Designator('pipeline_3'): PipelineResult(prediction=results_3,
                                                 bound_pipeline=BoundPipelineStub('pipeline_3')),
    })

    rankings = Rankings({'MinRanking': Ranking(metric=MetricMinVal(), results=pipeline_results)})

    identity_ranking = aggregator.aggregate(rankings=rankings)
    reverse_ranking = aggregator.aggregate(rankings=rankings, reverse=True)

    assert [s.pipeline_des for s in identity_ranking.scores(all_scores=True)] \
        == ['pipeline_1', 'pipeline_2', 'pipeline_3']
    assert [s.pipeline_des for s in reverse_ranking.scores(all_scores=True)] \
        == ['pipeline_3', 'pipeline_2', 'pipeline_1']


def test_fail_aggregator():
    '''Tests aggregators which raise exceptions.'''
    catalog = AggregatorCatalogAuto()
    catalog.register(TSADAggregator(
        name='test_aggregators.identity',
        tags={'for_tests': ['true']},
        fail=True,
    ))
    aggregator = catalog.lookup_by_name('test_aggregators.identity')

    results_1 = Dataset({
        '1': 1,
    })
    results_2 = Dataset({
        '1': 2,
    })
    results_3 = Dataset({
        '1': 3,
    })

    pipeline_results = PipelineResults({
        Designator('pipeline_1'): PipelineResult(prediction=results_1,
                                                 bound_pipeline=BoundPipelineStub('pipeline_1')),
        Designator('pipeline_2'): PipelineResult(prediction=results_2,
                                                 bound_pipeline=BoundPipelineStub('pipeline_2')),
        Designator('pipeline_3'): PipelineResult(prediction=results_3,
                                                 bound_pipeline=BoundPipelineStub('pipeline_3')),
    })

    rankings = Rankings({'MinRanking': Ranking(metric=MetricMinVal(), results=pipeline_results)})

    new_rankings = AggregateRanker(methods=[aggregator],
                                   rankings=rankings,
                                   results=pipeline_results,
                                   )()
    new_scores = new_rankings['test_aggregators.identity'].scores(all_scores=True)

    assert new_scores[0].score == 'intentional failure'
    assert new_scores[1].score == 'intentional failure'
    assert new_scores[2].score == 'intentional failure'



'''A connector model that changes the names in dictionaries.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict, List, Optional, Union

from ..algorithms.impl.algorithm import Algorithm, MemoryAlgorithmCatalog, InputKeyError
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..catalog.catalog import upcast
from ..wrangler.dataset import Dataset


class ConnectorModelInstance(AlgorithmInstance):
    '''Connect one model to another by transforming names.'''
    _transforms: Dict[str, List[str]]

    def __init__(self, parent, **hyperparams: Union[str, List[str]]):
        super().__init__(parent=parent)
        self._transforms = {}
        transforms = self.algorithm.hyperparams(**hyperparams)
        for new_key, old_path in transforms.items():
            if isinstance(old_path, str):
                self._transforms[new_key] = [old_path]
            else:
                self._transforms[new_key] = old_path

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        retval = dataset.output()
        for new_key, old_path in self._transforms.items():
            here = dataset
            for old_key in old_path:
                try:
                    there = here[old_key]
                except KeyError as exc:
                    raise InputKeyError(
                        f'Can not find old path "{".".join(old_path)}"'
                        f' to rename to new key "{new_key}".'
                        f' Looking for "{old_key}".'
                        f' Keys here are [{", ".join(list(here.keys()))}].') from exc
                here = there
            retval[new_key] = here

        return retval


class ConnectorModel(Algorithm):
    '''Connect one model to another by transforming names.
    Usage:
    connector = ConnectorModel()
    instance = connector.instantiate(newkey1='oldkey1', newkey2='oldkey2')
    old_dataset = {'oldkey1':'v1', 'oldkey2':'v2'}
    new_dataset = instance.predict(old_dataset)
    #new_dataset == {'newkey1':'v1', 'newkey2':'v2'}

    If you want to keep a key the same, pass it in as connector.instantiate(samekey='samekey')
    If you want to drop a key and its corresponding value, don't pass it in as a new key.

    You get only the keys that are mentioned in the arguments to instantiate.
    All other key/value pairs are discarded.
    '''
    _name = 'connect'
    _instance_constructor = ConnectorModelInstance
    _tags = {
        'source': ['auton_lab'],
        'preprocessor': ['true'],
    }


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = ConnectorModel(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Tests for execption_thread.py.'''

import pytest
from .exception_thread import ExceptionThread


def test_sunny_day() -> None:
    '''Test the ExceptionThread class.'''

    class TestException(Exception):
        '''Test exception class.'''

    def raising_function(should_raise: bool = False):
        '''Conditionally raise an exception.'''
        if should_raise:
            raise TestException('This is a test exception.')

    et = ExceptionThread(target=raising_function, kwargs={'should_raise': True})
    et.start()
    with pytest.raises(TestException, match='This is a test exception'):
        et.join()

    et2 = ExceptionThread(target=raising_function, kwargs={'should_raise': False})
    et2.start()
    et2.join()



'''Events that a NeighborManager can report.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


from typing import Any

from .node_id import NodeID


class Event():
    '''Base class NeighborManager events.'''
    _neighbor: NodeID

    def __init__(self, neighbor: NodeID):
        self._neighbor = neighbor

    @property
    def neighbor(self):
        '''The relevant neighbor originating the event.'''
        return self._neighbor

    def __str__(self) -> str:
        return f'Event {self.__class__.__name__} from {self.neighbor}'

    def __repr__(self) -> str:
        return f'<Event {self.__class__.__name__} from {self.neighbor}>'


class Recv(Event):
    '''A received payload from a neighbor.'''
    _payload: bytes

    def __init__(self, neighbor: NodeID, payload: bytes):
        super().__init__(neighbor)
        self._payload = payload

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Recv):
            return self.neighbor == other.neighbor and self.payload == other.payload
        return False

    @property
    def payload(self) -> bytes:
        '''The inbound payload.'''
        return self._payload


class NewNeighbor(Event):
    '''A new neighbor has arrived.'''


class NeighborTimeout(Event):
    '''A neighbor has not been seen in a while.'''



'''An algorithm for computing moments.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Optional

import numpy as np

from ..algorithms.impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ..algorithms.impl.fittable_algorithm_instance import FittableAlgorithmInstance
from ..catalog.catalog import upcast
from ..wrangler.dataset import Dataset


class MomentsInstance(FittableAlgorithmInstance):
    '''A model that computes the moments.'''

    _mu: np.ndarray
    _cov: np.ndarray

    def fit(self, dataset: Optional[Dataset]):
        '''Compute moments

        We call fit with the domain adaptation source dataset, and
        then predict with the domain adaptation target dataset.
        '''
        if dataset is None:
            return
        self._trained = True

        numeric = dataset.dataframe.select_dtypes(include=['number'])

        self._mu = np.mean(numeric, axis=0)
        self._cov = np.cov(numeric.T)

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        ''''We produce a dataset that only has a 'hyperparameters' field
        that contains the mean and covariance of the input dataset.
        '''
        return Dataset(hyperparams={
            'mean': self._mu,
            'covariance': self._cov,
        })


class Moments(Algorithm):
    '''A model that calculates the moments of a dataset.'''
    _name = 'moments'
    _instance_constructor = MomentsInstance
    _tags = {
        'source': ['auton_lab'],
        'preprocessor': ['true'],
    }


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = Moments(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''A table of data, supporting multiple formats.'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.
import abc
from collections import defaultdict
from typing import Any, Callable, Dict, List, Type, Union
import warnings

warnings.filterwarnings('ignore', category=FutureWarning)


class Error(BaseException):
    '''Base class for all table-related errors.'''
    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        comma_newline = ',\n'
        return f'{self.__class__.__name__}({comma_newline.join(self.args)})'


class CanNotInstantiate(Error):
    '''The implementaiton does not understand how to use the type of value.'''


class TableValueError(Error, ValueError):
    '''Raised when a table has the correct keys but invalid values.'''


class Table(metaclass=abc.ABCMeta):
    '''The interface for Table for one variety of Table.'''
    name: str
    tags: Dict[str, List[str]]
    my_type: Type  # This is the type of the value we are wrapping.
    _value: Any  # Has type my_type.

    # All subclasses must implement this __init__ signature.
    # When we get something from a catalog we don't otherwise know what
    # arguments it needs.
    def __init__(self, value: Any):
        if not isinstance(value, self.my_type):
            raise CanNotInstantiate()
        self._value = value

    def value(self):
        '''Get the native representation of the table.

        This is special because the return object is generally mutable.
        '''
        return self._value

    def __eq__(self, other: Any) -> bool:
        '''Test two objects for equality.'''
        if not isinstance(other, self.__class__):
            return False
        return False if self.my_type != other.my_type else self._value == other._value

    def __getitem__(self, key: Any) -> Any:
        '''Select a subitem of the table by key.'''
        return self._value[key]

    def __setitem__(self, key: Any, value: Any) -> None:
        '''Set a subitem of the table by key.'''
        self._value[key] = value

    def __len__(self) -> int:
        '''Return the length of this table.'''
        return len(self._value)

    def __str__(self) -> str:
        '''Return a string representation of the table.'''
        return f'Table({self._value})'

    def __repr__(self) -> str:
        '''Return a source representation of the table.'''
        return f'Table({self._value!r})'

    def copy(self) -> 'Table':
        '''Make a copy of the table.'''
        return self.__class__(self._value.copy())

    def equals(self, other: Any) -> bool:
        '''Test two objects for equality.'''
        if not isinstance(other, Table):
            return False
        return isinstance(other, self.__class__) and self._value.equals(other._value)  # pylint: disable=protected-access,line-too-long

    @property
    def shape(self) -> tuple:
        '''Get the shape of the table.'''
        return self._value.shape

    @abc.abstractmethod
    def drop(self, columns: List[Union[int, str]]) -> 'Table':
        '''Return a copy of this table with selected columns dropped.'''

    @property
    @abc.abstractmethod
    def columns(self) -> List[Union[int, str]]:
        '''Return the column names of the table.

        The can be passed to drop().
        '''

    @columns.setter
    @abc.abstractmethod
    def columns(self, value: List[Union[int, str]]) -> None:
        '''Set the column names of the table.'''

    @property
    @abc.abstractmethod
    def empty(self) -> bool:
        '''Return True if the table is empty.'''

    @abc.abstractmethod
    def to_csv(self, *args, **kwargs) -> str:
        '''Output the table as a CSV string.'''

    @abc.abstractmethod
    def as_dict(self, *args, **kwargs) -> Dict[Union[str, int], Any]:
        '''Convert table to a dictionary.'''


class TableFactory(Table):
    '''A table of data, supporting multiple formats.

    We only convert the data when a specific format is requested.

    Usage:

    TableFactory(foo) returns a FooTable

    e.g., if foo is a pandas dataframe, TableFactory(foo) returns a DataFrameTable containing foo.
    '''
    _exporters: Dict[type, Dict[type, Callable[['Table'], Any]]] = defaultdict(dict)
    _constructors: Dict[type, Callable[[Any], 'Table']] = {}
    # Ordered by issubclass() order, most general to most specific.
    _ordered_types: List[type] = []

    def __new__(cls, value: Any):
        '''Create a new table from a value.'''
        retval = None
        if type(value) in cls._constructors:
            retval = cls._constructors[type(value)](value)
        else:
            for my_type in cls._ordered_types:
                if isinstance(value, my_type):
                    retval = cls._constructors[my_type](value)
                    break
        if retval is None:
            raise CanNotInstantiate(
                f'Can not instantiate a Table from this {type(value)} value: {str(value)[:255]}')
        # TODO(Merritt): add ... to above error message if longer than 255 charas
        # Copy the as_() method from the class to the instance.
        # A better solution would be to make inheritance work properly, but I can't figure that out.
        setattr(retval, 'as_', cls.as_.__get__(retval))  # pylint: disable=no-value-for-parameter
        setattr(retval, '_cls', cls)
        return retval

    def as_(self, to_type: type) -> Any:
        '''Return the table as the requested type.'''
        try:
            return self._cls._exporters[self.my_type][to_type](self)  # type: ignore[attr-defined] # pylint: disable=no-member,protected-access
        except KeyError:
            raise TableValueError(f'Can not export {self.my_type} table to a {to_type}')  # pylint: disable=line-too-long,raise-missing-from

    @classmethod
    def register_constructor(
            cls, my_type: type, constructor: Callable[[Any], Table]) -> None:
        '''Register a new table type.

        Enables tables to be created with syntax like (e.g. using ``pd.DataFrame``
        and ``DataFrameTable``):

        .. code-block:: python

            my_df = pd.DataFrame()
            foo = Table(my_df)  # type: ignore[abstract]

        And the type of ``foo`` will be ``DataFrameTable``.

        '''
        cls._constructors[my_type] = constructor
        cls._ordered_types.append(my_type)

    @classmethod
    def register_exporter(
            cls,
            from_type: type, to_type: type,
            exporter: Callable[['Table'], Any]) -> None:
        '''Register a new exporter pair.

        Enables the as_() function:

        .. code-block: python

            foo = my_table.as_(pd.DataFrame).iloc[0, 2]

            my_np = my_table.as_(np.ndarray)
        '''
        cls._exporters[from_type][to_type] = exporter

    # Everything below this point is just here to make the linter happy.
    # (Otherwise, this class is abstract and it gives an error for instantiating an abstract class)
    # We don't actually use these methods in the codebase.

    def to_csv(self, *args, **kwargs) -> str:
        '''Output the table as a CSV string.'''
        return 'table_factory'

    def as_dict(self, *args, **kwargs) -> Dict[Union[str, int], Any]:
        '''Convert table to a dictionary.'''
        return {'table_factory': 'table_factory'}

    def drop(self, columns: List[Union[int, str]]) -> Table:
        '''Return a copy of this table with selected columns dropped.'''
        return self

    @property
    def columns(self) -> List[Union[int, str]]:
        '''Return the column names of the table.'''
        return []

    @columns.setter
    def columns(self, value: List[Union[int, str]]):
        '''Set the column names of the table.'''
        _ = value

    @property
    def empty(self):
        '''Return True if the table is empty.'''
        return True



'''Catalog that can be updated by plugins.'''

from typing import Callable, Optional, TypeVar
import os
from pathlib import Path
import sys

from ..wrangler.constants import PACKAGE_NAME
from .catalog import Catalog
from .catalog_shelf import CatalogShelf
from .memory_catalog import MemoryCatalog
from .pathed_catalog import PathedCatalog

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


T = TypeVar("T")


class PluginCatalog(CatalogShelf[T]):
    '''This catalog loads plugins for the given catalog_name.

    default_root is only needed for testing.

    We load the default auto catalog from ../{catalog_name},
    then prepend all the plugins,
    then prepend the home directory.
    '''
    def __init__(self, catalog_name: str, default_root: Optional[Path] = None, **kwargs):

        default_root = default_root or Path(__file__).parent
        super().__init__(PathedCatalog(paths=[default_root / catalog_name], **kwargs), **kwargs)

        group = f'{PACKAGE_NAME}.make_catalog'
        discovered_plugins = entry_points(group=group, name=catalog_name)
        for entrypoint in discovered_plugins:
            make_catalog: Callable[[], Catalog] = entrypoint.load()
            self.prepend(make_catalog(**kwargs))

        home_dir = os.getenv('HOME', '/')
        plugin_dir = Path(home_dir) / f'.{PACKAGE_NAME}' / 'plugins' / catalog_name
        self.prepend(PathedCatalog(paths=[plugin_dir], **kwargs))
        self.validate()
        # Add an empty catalog to take new registrations.
        self.prepend(MemoryCatalog(**kwargs))

    def validate(self):
        '''Apply constraints specific to the catalog type.'''



'''Tests for one_hot_encoder.'''
# pylint: disable=missing-function-docstring, duplicate-code, missing-class-docstring
# pylint: disable=redefined-outer-name

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Generator, Optional, Tuple

import pandas as pd
import pytest

from ....algorithms.impl.fittable_algorithm_instance import UntrainedError
from ....wrangler.dataset import Dataset, DatasetKeys

from .one_hot_encoder import OneHotModel, OneHotModelInstance


@pytest.fixture(autouse=True)
def model_and_data() -> Generator[Tuple[OneHotModelInstance, Dataset], None, None]:
    instance = OneHotModel().instantiate()

    # Simple example dataset from:
    # https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html

    students = [['Male', 1], ['Female', 3], ['Female', 2]]

    students_df = pd.DataFrame(students, columns=['gender', 'group'])

    students_dataset = Dataset({
        DatasetKeys.DATAFRAME.value: students_df})

    yield ((instance, students_dataset))


class TestOneHotEncoder:
    def test_sunny_day(self, model_and_data) -> None:
        instance, students_dataset = model_and_data
        assert isinstance(instance, OneHotModelInstance)
        assert isinstance(students_dataset, Dataset)

        instance.fit(students_dataset)
        result: Optional[Dataset] = instance.predict(students_dataset)

        assert result is not None
        students = [[1, 0.0, 1.0], [3, 1.0, 0.0], [2, 1.0, 0.0]]
        expected_result = pd.DataFrame(students, columns=['group', 'gender_Female', 'gender_Male'])

        assert result.dataframe.equals(expected_result)

    def test_untrained(self, model_and_data) -> None:
        instance, students_dataset = model_and_data
        assert isinstance(instance, OneHotModelInstance)
        assert isinstance(students_dataset, Dataset)

        with pytest.raises(UntrainedError):
            instance.predict(students_dataset)

    def test_unfamiliar_category(self, model_and_data) -> None:
        '''If the encoder encounters a category it wasn't fit on,
        we want it to put a 0 in all columns for that category instead of failing.'''
        instance, students_dataset = model_and_data
        assert isinstance(instance, OneHotModelInstance)
        assert isinstance(students_dataset, Dataset)

        instance.fit(students_dataset)

        df1: pd.DataFrame = students_dataset.dataframe
        df2 = pd.DataFrame({'gender': ['Non-binary'], 'group': [2]})
        students_dataset.dataframe = pd.concat([df1, df2], ignore_index=True)
        result: Optional[Dataset] = instance.predict(students_dataset)

        assert result is not None
        students = [
            [1, 0.0, 1.0],
            [3, 1.0, 0.0],
            [2, 1.0, 0.0],
            [2, 0.0, 0.0]]
        expected_result = pd.DataFrame(students, columns=['group', 'gender_Female', 'gender_Male'])
        assert result.dataframe.equals(expected_result)



'''A step consisting of parallel pipelines.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List, Optional

from ...algorithms.impl.algorithm import Algorithm, AlgorithmCatalog
from ...generator.designator import Designator
from .pipeline_step import PipelineStep, GeneratorInterface


class Error(Exception):
    '''Base error for ParallelStep'''


class GeneratorError(Error):
    '''Something has gone wrong during generate()'''


class ParallelStep(PipelineStep):
    '''A collection of parallel pipelines

    The output of each pipeline will be in a Dataset
    object keyed by the parameters given in the constructor.
    '''
    # This is really Dict[str, PipelineTemplate], but that's a
    # circular reference
    _subpipelines: Dict[str, Any]

    def __init__(self,
                 model: Optional[Algorithm] = None,
                 algorithm_catalog: Optional[AlgorithmCatalog] = None,
                 generator: Optional[GeneratorInterface] = None,
                 **pipelines):
        super().__init__(model=model, algorithm_catalog=algorithm_catalog, generator=generator)
        self._subpipelines = pipelines

    @property
    def subpipeline_keys(self) -> List[str]:
        '''Get the keys for all the pipelines.'''
        return list(self._subpipelines.keys())

    @property
    def subpipelines(self) -> Dict[str, Any]:
        '''Get all the subpipelines.

        Really returns a Dict[str, PipelineTemplate], but that's a
            circular reference.
        '''
        return self._subpipelines

    def subpipeline(self, key: str) -> Any:
        '''Look up an individual pipeline by the key given in the constructor.

        Args:
          key: the Dataset key that will be used to hold the results of the given
          pipeline.

        Returns:
          The PipelineTemplate keyed by key.
        '''
        return self._subpipelines[key]

    def lift(self) -> 'PipelineStep':
        '''View a ParallelStep as just a PipelineStep.'''
        return self

    def generate(self,
                 future_steps: List['PipelineStep']
                 ) -> Dict[Designator, List['PipelineStep']]:
        '''Generate pipeline sketches (lists of bound steps) for this step and all subsequent steps.

        Subsequent query steps may expand into multiple 'pipeline sketches'. We do not currently
        support query steps within parallel pipelines.

        ParallelStep version.

        Our designator component looks like

            my_name(p1=a_step:another_step,p2=different_step)

        where p1 and p2 are the keys to the parallel constructor in asciibetical order.
        '''
        assert self._generator is not None, (
            'BUG: PipelineStep built without a generator, '
            f'name: {self.pipeline_designator_component}')
        new_subpipelines: Dict[str, Dict[Designator, Any]] = {
            key: self._generator.generate(pipeline)
            for key, pipeline in self._subpipelines.items()
        }

        # Calculating an n-way cross product is hard, so for
        # the moment, we do not try.
        # Real type is Dict[str, BoundPipeline].
        bound_subpipelines: Dict[str, Any] = {}
        long_designators: List[Designator] = []
        for key, candidate_pipelines in new_subpipelines.items():
            # If there is more than 1 candidate_pipeline, there must
            # have been a query in the subpipeline template, which is currently
            # unsupported.
            if len(candidate_pipelines) != 1:
                raise GeneratorError(
                    f'{key}: Subpipelines that multiply are not currently supported.'
                )
            # Extract the unique subpipeline.
            bound_subpipelines[key] = list(candidate_pipelines.values())[0]
            subkey = list(candidate_pipelines.keys())[0]
            long_designators.append(Designator(f'{key}={subkey}'))

        new_step = self.__class__(
            model=None,
            algorithm_catalog=self._algorithm_catalog,
            generator=self._generator,
            **bound_subpipelines).lift()

        if len(future_steps) == 0:
            return {Designator('_'): [new_step]}
        step0 = future_steps[0]
        sketches = step0.generate(future_steps[1:])
        return {
            Designator(f'{key}'): [new_step] + rest[:]
            for key, rest in sketches.items()
        }



'''Tests for metric_config'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pytest

from .metric_config import MetricConfig, MetricConfigError
# pylint: disable=missing-function-docstring,duplicate-code


def test_metric_config_rainy_day():
    dut = MetricConfig(clause={
        'm1': {'param1': 'value1'},
        'm2': {},
    })
    with pytest.raises(MetricConfigError, match='must have exactly 1'):
        dut.validate()



'''Tests for matchers.py'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Optional
from ..algorithms.impl.algorithm import Algorithm
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..generator.bound_pipeline import BoundPipeline
from ..wrangler.constants import Matcher
from ..wrangler.dataset import Dataset

from .matchers import MatcherFactory

# pylint: disable=missing-class-docstring, missing-function-docstring, duplicate-code


class FakeAlgorithmInstance(AlgorithmInstance):
    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        return Dataset(dataset.copy())


class FakeAlgorithm(Algorithm):
    _name = 'fake_algorithm'
    _default_hyperparams = {
        'base_param': 'base_value',
        'param_to_override': 'value_to_override',
    }
    _tags = {
        'some_tag': [
            'some_tag_value',
            'different_tag_value',
        ],
        'another_tag': [
            'another_tag_value',
        ]
    }

    def instantiate(self, **hyperparams) -> FakeAlgorithmInstance:
        return FakeAlgorithmInstance(parent=self, **hyperparams)


def test_algorithm_sunny_day():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step(model=FakeAlgorithm())

    factory = MatcherFactory()
    dut = factory.make(Matcher.ALGORITHM, args='fake_algorithm')
    assert dut.matches(step, pipeline)


def test_algorithm_no_match():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step(model=FakeAlgorithm())

    factory = MatcherFactory()
    dut = factory.make(Matcher.ALGORITHM, args='real_algorithm')
    assert not dut.matches(step, pipeline)


def test_algorithm_no_model():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step().set_name('some_name')

    factory = MatcherFactory()
    dut = factory.make(Matcher.ALGORITHM, args='some_name')
    assert not dut.matches(step, pipeline)


def test_designator_sunny_day():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step(model=FakeAlgorithm()).mark_queried()

    factory = MatcherFactory()
    dut = factory.make(Matcher.DESIGNATOR, args='a_pipeline@fake_algorithm')
    assert dut.matches(step, pipeline)


def test_designator_no_model():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step()

    factory = MatcherFactory()
    dut = factory.make(Matcher.DESIGNATOR, args='a_pipeline')
    assert dut.matches(step, pipeline)


def test_designator_no_match():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step(model=FakeAlgorithm())

    factory = MatcherFactory()
    dut = factory.make(Matcher.DESIGNATOR, args='fake_algorithm')
    assert not dut.matches(step, pipeline)


def test_name_sunny_day():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step(model=FakeAlgorithm()).set_name('fake_step')

    factory = MatcherFactory()
    dut = factory.make(Matcher.NAME, args='fake_step')
    assert dut.matches(step, pipeline)


def test_name_no_match():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step(model=FakeAlgorithm()).set_name('fake_step')

    factory = MatcherFactory()
    dut = factory.make(Matcher.NAME, args='real_step')
    assert not dut.matches(step, pipeline)


def test_name_no_model():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step().set_name('fake_step')

    factory = MatcherFactory()
    dut = factory.make(Matcher.NAME, args='fake_step')
    assert dut.matches(step, pipeline)


def test_tags_sunny_day():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step(model=FakeAlgorithm())

    factory = MatcherFactory()
    dut = factory.make(Matcher.TAGS, args={
        'some_tag': 'some_tag_value',
    })
    assert dut.matches(step, pipeline)


def test_tags_no_match():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step(model=FakeAlgorithm())

    factory = MatcherFactory()
    dut = factory.make(Matcher.TAGS, args={
        'some_tag': 'nonmatching_tag_value',
    })
    assert not dut.matches(step, pipeline)


def test_tags_no_model():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step()

    factory = MatcherFactory()
    dut = factory.make(Matcher.TAGS, args={
        'some_tag': 'some_tag_value',
    })
    assert not dut.matches(step, pipeline)


def test_tags_all_must_match_failure():
    '''All the keys must have a matching value.'''
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step(FakeAlgorithm())

    factory = MatcherFactory()
    dut = factory.make(Matcher.TAGS, args={
        'some_tag': 'some_tag_value',
        'another_tag': 'missing_tag_value',
    })
    assert not dut.matches(step, pipeline)


def test_tags_all_must_match_sunny():
    pipeline = BoundPipeline(name='a_pipeline')
    step = pipeline.step(FakeAlgorithm())

    factory = MatcherFactory()
    dut = factory.make(Matcher.TAGS, args={
        'some_tag': 'some_tag_value',
        'another_tag': 'another_tag_value',
    })
    assert dut.matches(step, pipeline)



'''Base class for distributed alogorithm instances.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
import collections
from copy import deepcopy
from pathlib import Path
import pickle
import queue
import threading
import time
import traceback
from typing import (Any, Callable, Dict, Generic, Iterable, List, Mapping,
                    Optional, Set, Tuple, Type, TypeVar)

import numpy as np
from numpy.random import RandomState  # pylint: disable=no-name-in-module  # This just isn't true.
import pandas as pd

from ...algorithms.impl.fittable_algorithm_instance import UntrainedError
from ...algorithms.impl.neighbor_metadata import NeighborMetadata
from ...algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ...algorithms.impl.neighbor_state_interface import NeighborStateInterface
from ...config_components.distributed_config import (
    DistributedConfig, DropperConfig, CommunicatorConfig)
from ...deciders.impl.decider import Decider
from ...deciders.impl.decider_auto import DeciderCatalogAuto
from ...neighbor_manager.event import Recv
from ...neighbor_manager.node_id import NodeID
from ...neighbor_manager.neighbor_manager import NeighborManager
from ...neighbor_manager.neighbor_manager_impl import NeighborManagerImpl
from ...wrangler.dataset import Dataset, DatasetKeys
from ...wrangler.exception_thread import ExceptionThread
from ...wrangler.logger import Logger, Level

from ..impl.algorithm import Algorithm
from ..impl.fittable_algorithm_instance import FittableAlgorithmInstance


log = Logger(__file__, to_stdout=True, level=Level.INFO).logger()


NeighborStateSubclass = TypeVar('NeighborStateSubclass', bound='NeighborState')


class NoDataError(Exception):
    '''No data was provided to the algorithm.'''


class ForceFalse(Exception):
    '''Force the decider to return False.'''


class Dropper():
    '''A class that can be used to drop messages.'''
    _config: DropperConfig
    _tried_counter: int = 0
    _dropped_counter: int = 0

    def __init__(self, config: DropperConfig) -> None:
        self._config = config
        self._random = RandomState(seed=config.seed)
        self._first = True
        if config.output_dir is not None:
            log.info('Dropper output dir: %s', config.output_dir)
            config.output_dir.mkdir(parents=True, exist_ok=True)

    def drop(self) -> bool:
        '''Return True if the message should be dropped.'''
        retval = (not self._first) and (1 == self._random.binomial(n=1, p=self._config.drop_rate))

        self._first = False

        self._tried_counter += 1
        if retval:
            self._dropped_counter += 1
        return retval

    @property
    def output_dir(self) -> Optional[Path]:
        '''Return the output directory.'''
        return self._config.output_dir

    @property
    def tried_counter(self) -> int:
        '''Return the number of messages tried.'''
        return self._tried_counter

    @property
    def dropped_counter(self) -> int:
        '''Return the number of messages dropped.'''
        return self._dropped_counter


class NeighborState(
        Generic[NeighborStateSubclass],
        NeighborStateInterface[NeighborStateSubclass]):
    '''A message from a neighbor in the distributed network.'''
    _columns: Optional[List[str]]
    # TODO(Merritt): add a respesentation of total amount of data

    def __init__(self, columns: Optional[List[str]] = None) -> None:
        if columns is not None:
            # This is here because mypy is missing typing issues on the columns.
            columns = self._cast_columns(columns=columns)
        self._columns = columns

    def __str__(self) -> str:
        return f'{self.__class__.__name__}(columns = {self.columns})'

    def __eq__(self, value: Any) -> bool:
        if not isinstance(value, NeighborState):
            return False
        return self.columns == value.columns

    @property
    def columns(self) -> Optional[List[str]]:
        '''The column names.'''
        return self._columns

    @columns.setter
    def columns(self, value: Optional[List[str]]) -> None:
        '''The column names.'''
        if value is not None:
            # This is here because mypy is missing typing issues on the columns.
            value = self._cast_columns(columns=value)
        self._columns = value

    @abc.abstractmethod
    def encode(self) -> bytes:
        '''Encode a message for distributed neighbors. '''

    @classmethod
    @abc.abstractmethod
    def decode(cls, serialized_model: bytes) -> NeighborStateSubclass:
        '''Decode a message from distributed neighbors.

        This should just redirect to the relevant NeighborState subclass.
        '''

    @property
    def payload_size(self):
        '''Return the size of the payload to send this state.'''
        return self.encode().__sizeof__()

    def state_differs(self,
                      distributed: DistributedConfig,
                      other: Optional[NeighborStateSubclass]) -> bool:
        '''Decide if our state differs meaningfully from another.

        :distributed: contains information from the problem definition, including fit_eps
        :other: is an instance of type(self) or None.
        '''
        if other is None:
            return True

        state_distance = self.distance(other)
        log.log(Level.VERBOSE, 'node %s: state_distance: %s', distributed.my_id, state_distance)

        changed = bool(state_distance > distributed.fit_eps)

        return changed

    @abc.abstractmethod
    def distance(self, other: NeighborStateSubclass) -> float:
        '''A numerical measure of the distance between this state and another.'''

    @classmethod
    def _cast_columns(cls, columns: Any) -> List[str]:
        '''Return columns typed as a list of strings.

        Useful for type safety in subclasses' decode methods.
        '''
        assert isinstance(columns, collections.abc.Iterable), (
            'BUG: expected columns to be iterable, '
            f'instead found {columns} of type {type(columns)}'
        )
        columns = list(columns)
        assert all(isinstance(col_name, (int, str)) for col_name in columns), (
            f'BUG: expected column names to be ints or strings, instead found {columns} '
            f'of types {[type(col_name) for col_name in columns]}.'
        )
        return [str(c) for c in columns]


class DistributedAlgorithmInstance(FittableAlgorithmInstance):  # pylint: disable=too-many-public-methods,line-too-long
    '''Base class for distributed algorithm instances.'''

    _known_neighbors: Set[NodeID]
    _my_id: NodeID
    _neighbor_metadata: Dict[NodeID, NeighborMetadata]
    _my_state_metadata: NeighborMetadata
    _predict_state: Optional[NeighborStateInterface] = None
    _neighbor_constructor: Type[NeighborStateInterface]

    _lock_model: Optional[threading.Lock] = None
    _lock_predict_state: Optional[threading.Lock] = None
    _neighbor_manager: Optional[NeighborManager] = None
    _fit_thread: Optional[ExceptionThread] = None
    _read_thread: Optional[ExceptionThread] = None
    _exception_queue: Optional[queue.Queue[Tuple[Exception, str]]]
    _distributed: DistributedConfig
    _synchronous: bool = False  # Do we run our fit thread or wait to be told to fit?
    _dropper: Optional[Dropper] = None

    _convergence_count: int  # Number of times we considered fitting since the model last changed
    _changed_on_fit: bool  # Set to true if we decided to send last time we fit.
    _convergence_time: Optional[int]  # value of time.monotonic_ns() at start of most recent
    #                                   convergence period (None if we sent on last fit)
    _stop: bool = False
    _pending_message_from_neighbor: bool = False
    _training: bool = False   # True if we are currently training

    _last_dataset: Optional[Dataset] = None
    _columns: Optional[List[str]] = None
    _lambda: np.float64  # How strongly we weight our neighbors' models.
    _omega: np.float64  # How strongly we weight previous states of our own model.

    _fit_counter: int = 0
    _fit_periodic_counter: int = 0
    _predict_counter: int = 0
    _emit_counter: int = 0

    @classmethod
    def instantiate_with_state(
            cls,
            state: bytes,
            parent: Algorithm,
            distributed: DistributedConfig) -> 'DistributedAlgorithmInstance':
        '''Create an instance of this class from a saved state.'''
        instance = cls(parent=parent, distributed=distributed)
        instance._trained = True
        my_state = instance._neighbor_constructor.decode(state)
        instance._my_state_metadata = NeighborMetadata(my_state=my_state)
        instance._predict_state = my_state
        instance._lock_predict_state = threading.Lock()
        return instance

    @property
    def trained(self) -> bool:
        '''Is this algorithm trained?

        Override base class to take the lock.
        '''
        if self._lock_model is None:
            # We are in a deepcopy.
            return self._trained

        with self._lock_model:
            retval = self._trained
        return retval

    def __deepcopy__(self, memo: Dict[int, object]) -> 'DistributedAlgorithmInstance':
        '''Deepcopy implementation.'''
        # We don't want to deepcopy the lock, threads, the neighbor manager, the
        # last dataset, pending message state, the stop flag, or the exception queue.
        alg_name = self._algorithm.name if self._algorithm is not None else 'unknown'
        log.debug('%s: node %d (%x) is deepcopying.', alg_name, self._my_id, id(self))
        assert self._algorithm is not None
        retval = type(self)(parent=self._algorithm,
                            distributed=self._distributed,
                            Lambda=self._lambda,
                            omega=self._omega)
        retval._known_neighbors = deepcopy(self._known_neighbors, memo=memo)
        retval._my_id = self._my_id
        retval._neighbor_metadata = deepcopy(self._neighbor_metadata, memo=memo)
        # retval._my_state = deepcopy(self._my_state, memo=memo)
        retval._my_state_metadata = deepcopy(self._my_state_metadata, memo=memo)
        retval._predict_state = deepcopy(self._predict_state, memo=memo)

        retval._lock_model = None
        retval._lock_predict_state = None
        retval._exception_queue = None
        retval._distributed = deepcopy(self._distributed, memo=memo)

        retval._last_dataset = None
        if self._columns is not None:
            retval._columns = self._columns.copy()

        retval._pending_message_from_neighbor = False
        retval._trained = self._trained

        retval._convergence_time = self._convergence_time
        return retval

    def __init__(self,
                 parent: Algorithm,
                 distributed: DistributedConfig,
                 **kwargs):
        self._lambda = np.float64(kwargs.pop('Lambda'))
        self._omega = np.float64(kwargs.pop('omega'))
        self._tol = np.float64(kwargs.pop('tol'))
        _maxiter = kwargs.pop('maxiter')
        self._maxiter = int(_maxiter) if _maxiter is not None else None
        self._synchronous = bool(kwargs.pop('synchronous', False))
        super().__init__(parent, **kwargs)
        self._distributed = distributed
        self._my_id = distributed.my_id
        self._neighbor_metadata = {}
        self._my_state_metadata = NeighborMetadata()
        self._known_neighbors = set()
        self._lock_model = threading.Lock()
        self._lock_predict_state = threading.Lock()
        self._convergence_count = 0
        self._convergence_time = None
        self._exception_queue = queue.Queue()
        self._changed_on_fit = False
        if self._distributed.dropper is not None:
            self._dropper = Dropper(config=self._distributed.dropper)

    @property
    def _neighbor_models_iter(self) -> Iterable[NeighborState]:
        for _, v in self._neighbor_metadata.items():
            if v.current_state is None:  # This is a neighbor we haven't heard from yet.
                continue
            assert isinstance(v.current_state, NeighborState), (
                'BUG: expected neighbor_models to contain NeighborState, '
                f'instead found {v.current_state} of type {type(v.current_state)}'
            )
            yield v.current_state

    @property  # type: ignore[override]
    def _my_state(self) -> Optional[NeighborStateInterface]:
        '''Get the state of this node.'''
        return self._my_state_metadata.current_state

    @_my_state.setter
    def _my_state(self, value: Optional[NeighborStateInterface]) -> None:
        self._my_state_metadata.current_state = value

    def decide(self,
               neighbor_id: NodeID,
               neighbors: Optional[Mapping[NodeID, NeighborMetadataInterface]] = None) -> bool:
        '''Decide if we should send an update to a specific neighbor.

        :distributed: contains information from the problem definition.
        :neighbor_id: is the ID for the node we want to decide against.
        :neighbors: is an optional mapping of all neighbors. If None, we
            use self._neighbor_metadata. This argument is to simplify testing.
        '''
        results: List[bool] = []
        decider_catalog = DeciderCatalogAuto()
        if neighbors is None:
            neighbors = self._neighbor_metadata
        try:
            for name, config in self._distributed.decider.decider_configs.items():
                if config.enabled:
                    decider: Decider = decider_catalog.lookup_by_name(
                        name)(name=name, config=config)
                    results.append(decider.decide(
                        my_meta=self._my_state_metadata,
                        neighbor_id=neighbor_id,
                        neighbors=neighbors))
        except ForceFalse:
            return False
        return any(results)

    @property
    def locked(self) -> bool:
        '''Check the model locks status'''
        return False if self._lock_model is None else self._lock_model.locked()

    @property
    def converged(self) -> bool:
        '''Check if the model has converged.'''
        if self._lock_model is None:
            # We are in a deepcopy.
            return self._converged()
        with self._lock_model:
            retval = self._converged()
        return retval

    def _converged(self) -> bool:
        '''Check if the model has converged.'''
        if self._convergence_time is None:
            log.debug('node %s (%x) has not converged.', self._my_id, id(self))
            return False
        now = time.monotonic_ns()
        elapsed = now - self._convergence_time
        if elapsed > self._distributed.convergence_time_ns:
            log.info('node %s (%x) has converged.', self._my_id, id(self))
        else:
            log.debug('node %s (%x) has been converging for %s ns.', self._my_id, id(self), elapsed)
        return elapsed > self._distributed.convergence_time_ns

    @property
    def my_id(self) -> NodeID:
        '''Get the ID of this node.'''
        return self._my_id

    @property
    def my_state(self) -> Optional[NeighborStateInterface]:
        '''Get the state of this node.'''
        return self._my_state

    @property
    def predict_state(self) -> Optional[NeighborStateInterface]:
        '''Get the state of this node for prediction.'''
        return self._predict_state

    @property
    def my_state_copy(self) -> Optional[NeighborStateInterface]:
        '''Safely get a copy of the state of this node.'''
        if self._lock_model is None:
            # We are in a deepcopy and can not change.
            return self._my_state

        with self._lock_model:
            retval = deepcopy(self._my_state)
        return retval

    @property
    def dropper(self) -> Optional[Dropper]:
        '''Get the dropper.'''
        return self._dropper

    def start(self) -> None:
        '''Start all supporting threads.'''
        alg_name = self._algorithm.name if self._algorithm is not None else 'unknown'
        log.info('%s: node %s (%x) is starting', alg_name, self._my_id, id(self))
        self._stop = False
        if self._neighbor_manager is None:
            self._neighbor_manager = self._init_neighbor_manager()
            self._neighbor_manager.start()

        if not self._synchronous:
            self._fit_thread = ExceptionThread(target=self._fit_periodically, kwargs={
                'stop': lambda: self._stop
            })
            self._fit_thread.start()
            self._read_thread = ExceptionThread(target=self._read_from_neighbors, kwargs={
                'stop': lambda: self._stop
            })
            self._read_thread.start()

    @property
    def synchronous(self) -> bool:
        '''Return whether we are running synchronously.'''
        return self._synchronous

    def poll_exceptions(self) -> Optional[Tuple[Exception, str]]:
        '''Return most recent exception from inside a thread, if any.'''
        assert self._exception_queue is not None, (
            'BUG: _exception_queue should be set in __init__.  Is this a deepcopy?'
        )
        try:
            return self._exception_queue.get(block=False)
        except queue.Empty:
            return None

    def stop(self) -> None:
        '''Stop all supporting threads.'''
        alg_name = self._algorithm.name if self._algorithm is not None else 'unknown'
        log.info('%s: node %s (%x) is stopping.', alg_name, self._my_id, id(self))
        self._stop = True
        if self._fit_thread is not None:
            self._fit_thread.join()
        if self._read_thread is not None:
            self._read_thread.join()
        if self._neighbor_manager is not None:
            self._neighbor_manager.stop()

    def _init_neighbor_manager(self) -> NeighborManager:
        comm = self.algorithm.communicator_catalog.lookup_by_name(
            self._distributed.communicator.name)
        disc = self.algorithm.discoverer_catalog.lookup_by_name(
            self._distributed.discoverer)
        communicator = comm(distributed=self._distributed,
                            known_neighbors=self._known_neighbors,
                            my_id=self._my_id)
        discoverer = disc(config=self._distributed, communicator=communicator)
        return NeighborManagerImpl(discoverer=discoverer)

    def _read_from_neighbors(self, stop: Callable[[], bool]) -> None:
        '''Read from neighbors and update the neighbor states.'''
        log.debug('node %s (%x) is starting to receive from neighbors.', self._my_id, id(self))
        while not stop():
            self.read_from_neighbors()
            time.sleep(self._distributed.polling_interval)

    def read_from_neighbors(self) -> None:
        '''Read from neighbors and update our local neighbor states.'''
        try:
            if self._neighbor_manager is not None:
                for event in self._neighbor_manager.poll_for_events(
                        timeout=self._distributed.polling_interval):
                    log.log(Level.VERBOSE,
                            'node %s (%x) is reading %s', self._my_id, id(self), event)
                    if isinstance(event, Recv):
                        self._recv(event.neighbor, event.payload)
        except Exception as e:  # pylint: disable=broad-exception-caught
            trace = traceback.format_exc()
            log.error(
                'node %s (%x) error while reading from neighobrs:\n%s',
                self._my_id, id(self), trace)
            assert self._exception_queue is not None, (
                'BUG: _exception_queue should have been set in __init__. Is this a deepcopy?')
            self._exception_queue.put((e, trace))

    def _fit_periodically(self, stop: Callable[[], bool]):
        '''Refit if we have seen a message from a neighbor and enough time passed.'''
        log.debug('node %s (%x) is starting to fit periodically.', self._my_id, id(self))
        while not stop():
            assert self._lock_model is not None, (
                'BUG: _lock_model should have been set in __init__. Is this a deepcopy?'
            )

            self.fit_now()
            time.sleep(self._distributed.polling_interval)

    def fit_now(self) -> None:
        '''Fit the model now.

        This is the method that _fit_periodically calls to fit the model.

        If we are running in synchronous mode, we need to call this method to advance the model.
        '''
        assert self._lock_model is not None, (
            'BUG: _lock_model should have been set in __init__. Is this a deepcopy?')
        with self._lock_model:
            did_change = self._changed_on_fit

        if self._pending_message_from_neighbor or did_change:
            log.info(
                'node %s (%x) is spontaneously fitting. '
                'Pending message: %s. changed on last fit: %s.',
                self._my_id,
                id(self),
                self._pending_message_from_neighbor,
                did_change)
            try:
                self._fit_periodic_counter += 1
                self.fit(self._last_dataset)
            except Exception as e:  # pylint: disable=broad-exception-caught
                trace = traceback.format_exc()
                log.error(
                    'node %s (%x) error spontaneously fitting:\n%s',
                    self._my_id, id(self), trace)
                assert self._exception_queue is not None, (
                    'BUG: exception queue should have been set in __init__.  '
                    'Is this a deepcopy?')
                self._exception_queue.put((e, trace))
        if self._my_state is not None:
            with self._lock_model:
                self._emit_fit_complete()
                self._conditionally_send()

    @property
    def fit_counter(self) -> int:
        '''Return the number of times the model has been fit.'''
        return self._fit_counter

    @property
    def fit_periodic_counter(self) -> int:
        '''Return the number of times the model has been fit spontaneously.'''
        return self._fit_periodic_counter

    @property
    def predict_counter(self) -> int:
        '''Return the number of times the model has been predicted.'''
        return self._predict_counter

    def _recv(self, neighbor_id: NodeID, serialized_model: bytes) -> None:
        '''Receive a message from a neighbor.'''
        if self.dropper is not None and self.dropper.drop():
            return
        model = self._decode(serialized_model)
        assert self._lock_model is not None
        with self._lock_model:
            log.info('node %s (%x) received from %s', self._my_id, id(self), neighbor_id)
            log.log(Level.VERBOSE, 'received message: \n %s', model)
            self._pending_message_from_neighbor = True
            if neighbor_id not in self._neighbor_metadata:
                self._neighbor_metadata[neighbor_id] = NeighborMetadata()
            self._neighbor_metadata[neighbor_id].current_state = model

    @abc.abstractmethod
    def _decode(self, serialized_model: bytes) -> NeighborState:
        '''Decode a message from distributed neighbors. '''

    def _send(self) -> None:
        '''Send the model to all neighbors.

        Must be called with the model lock held.
        '''
        log.info('node %s (%x) is sending', self._my_id, id(self))
        assert self._my_state is not None, (
            f'node {self._my_id} ({id(self)})) has no state to send.')
        if self._neighbor_manager is not None:
            log.log(Level.VERBOSE, 'node %s (%x) has a neighbor manager.', self._my_id, id(self))
            self._neighbor_manager.send_all(self._my_state.encode())
        else:
            log.warning('node %s (%x) has no neighbor manager.', self._my_id, id(self))
            return
        # Only reset the last_state_sent if we actually sent a message.
        # We want to catch cumulative drift in self._my_state.v.
        log.debug("node %s (%x) reset last_state_sent.", self._my_id, id(self))
        self._my_state_metadata.sent_current_state()

    def _update_columns(self, dataset: Dataset) -> Dataset:
        '''set self._columns based on the given dataset and cross-check with neighbors.

        If neighbors are reporting a column that we don't see, call
            _resolve_neighbor_column_mismatch.
        '''

        self._columns = self._get_columns(dataset=dataset)

        mismatched_neighbors = {
            node_id: set(neighbor.current_state.columns)
            for node_id, neighbor in self._neighbor_metadata.items()
            if (neighbor.current_state is not None
                and neighbor.current_state.columns is not None
                and not set(neighbor.current_state.columns) == set(self._columns))
        }
        if mismatched_neighbors:
            return self._resolve_neighbor_column_mismatch(dataset, mismatched_neighbors)

        return dataset

    def _resolve_neighbor_column_mismatch(self,
                                          dataset: Dataset,
                                          mismatched_neighbors: Dict[NodeID, Set[str]]) -> Dataset:
        '''Handle case where set of columns reported by a neighbor does not match our own.

        Message contains information about
        '''
        _ = dataset
        assert self._columns is not None, (
            'BUG: self._columns should be set before _resolve_neighbor_column_mismatch is called.'
        )
        message = (
            'Found neighbors with mismatched columns: \n'
            + '\n'.join(f'Neighbor node {node_id} has columns {cols}.'
                        for node_id, cols in mismatched_neighbors.items())
            + f'\n columns we see in the data: {self._columns}. \n'
        )
        log.error(message)
        raise NotImplementedError(message)

    def _get_columns(self, dataset: Dataset) -> List[str]:
        '''Get the columns of the dataset.'''
        if DatasetKeys.DATAFRAME.value in dataset:
            # TODO(piggy): Decide if we need to remove the target column.
            retval = dataset.dataframe.columns
        elif DatasetKeys.COVARIATES.value in dataset:
            retval = dataset[DatasetKeys.COVARIATES.value].columns
        else:
            raise NoDataError('No data found in dataset.')
        return [str(c) for c in retval]

    def _is_empty(self, dataset: Dataset) -> bool:
        '''Return True if a dataset cannot be used to fit/predict.'''
        keys_to_check = [DatasetKeys.DATAFRAME.value, DatasetKeys.COVARIATES.value]
        for k in keys_to_check:
            if k in dataset:
                df_k = pd.DataFrame()
                try:
                    df_k = pd.DataFrame(dataset[k])
                    if df_k.shape[1] > 0:  # at least one column
                        return False  # Found non-empty dataframe.
                except Exception as err:  # pylint: disable=broad-exception-caught
                    log.error('Could not transform item at key %s to dataframe: %s\n',
                              k, err)
        return True  # Found no non-empty dataframes.

    def _emit_fit_complete(self) -> None:
        '''Emit a message when the model has been fit.'''
        if self._dropper is None:
            return
        self._emit_counter += 1
        record = {
            'my_id': self._my_id,
            'now': time.monotonic_ns(),
            'emit_counter': self._emit_counter,
            'fit_counter': self._fit_counter,
            'fit_periodic_counter': self._fit_periodic_counter,
            'predict_counter': self._predict_counter,
            'dropper_tried_counter': self._dropper.tried_counter,
            'dropper_dropped_counter': self._dropper.dropped_counter,
            'my_state': self._my_state.encode() if self._my_state is not None else None,
        }
        output_dir = self._dropper.output_dir
        if output_dir is not None:
            with open(output_dir
                      / f'id{self._my_id:02d}_{self._emit_counter:010d}_fit_complete.pkl', 'wb'
                      ) as f:
                f.write(pickle.dumps(record))

    def fit(self, dataset: Optional[Dataset], **kwargs) -> None:
        '''Fit the model to the data.

        If dataset is None, we process information from neighbors.
        If dataset is not None but is empty, we ignore it. This allows
        us to work with special data loaders.
        '''
        log.info('node %s (%x) is fitting (%s).',
                 self._my_id,
                 id(self),
                 'no data' if dataset is None else 'with data')

        assert self._lock_model is not None
        with self._lock_model:

            # Set this now so it will be False if an error interrupts the fit.
            self._changed_on_fit = False

            # Clear neighbor message now so we don't refit forever if a messge triggers an error
            self._pending_message_from_neighbor = False

            if dataset is not None:
                if self._is_empty(dataset):
                    log.warning('Node %s encountered empty or malformed dataset on fit: \n%s\n',
                                self._my_id, dataset)
                else:
                    dataset = self._update_columns(dataset.sorted_columns())
                    self._last_dataset = dataset
            self._training = True
            self._fit(dataset, **kwargs)
            self._training = False

            # Set after fit so we are not marked trained if we run into an error
            self._trained = True
            self._fit_counter += 1  # TODO(piggy): Move to self._my_state_metadata

            log.debug(
                'node %s (%x) (%s) fit complete; state: %s',
                self._my_id,
                id(self),
                self.__class__.__name__,
                self._my_state
            )
            log.log(
                Level.VERBOSE,
                'full state: %r', self._my_state)

            if self._my_state is not None:
                if not self._distributed.only_send_periodically:
                    self._conditionally_send()
                else:
                    log.info('node %s is not sending.', self._my_id)

        self._update_predict_state()

    def _detect_convergence(self, state_changed: bool) -> None:
        '''Detect if the model has converged.'''
        if self._my_state is None:
            return
        neighbor_distances = [
            self._my_state.distance(n) for n in self._neighbor_models_iter
        ]
        log.info('node %s neighbor distances: %s', self._my_id, neighbor_distances)
        neighbors_differ = [d > self._distributed.fit_eps for d in neighbor_distances]

        # State changed and not first fit
        state_meaningfully_changed = (
            state_changed
            and self._my_state_metadata.last_state_sent is not None)

        if state_meaningfully_changed and any(neighbors_differ):
            log.debug('node %s is non-converged.', self._my_id)
            # Disable convergence clock
            self._convergence_time = None
        else:
            # State did not change and/or first fit and/or matches all neighbors
            # Start convergence clock
            if self._convergence_time is None:
                log.debug('node %s is starting convergence clock.', self._my_id)
                self._convergence_time = time.monotonic_ns()
            convergence_duration_s = 1e9 * (time.monotonic_ns() - self._convergence_time)
            log.debug('node %s is converged for %s seconds.',
                      self._my_id,
                      convergence_duration_s)

    def _update_predict_state(self):
        '''Update the predict state with the current state.'''
        assert self._lock_predict_state is not None, (
            'BUG: _lock_predict_state should have been set in __init__. Is this a deepcopy?'
        )
        if self._my_state is not None:
            with self._lock_predict_state:
                self._predict_state = self.my_state_copy

    def _record_current_state_sent(self, positive_decisions: List[NodeID]) -> None:
        for neighbor_id in positive_decisions:
            # We might have not heard from this neighbor yet.
            if neighbor_id not in self._neighbor_metadata:
                self._neighbor_metadata[neighbor_id] = NeighborMetadata()
            self._neighbor_metadata[neighbor_id].last_state_sent = self._my_state

    def _send_broadcast_no_decisions(self, state_changed: bool) -> None:
        if state_changed:
            log.info('node %s (%x) is broadcasting now due to state change.',
                     self._my_id, id(self))
            self._send()

    def _send_unicast_no_decisions(self) -> None:
        assert self._my_state is not None
        assert self._neighbor_manager is not None
        positive_decisions = []
        for neighbor_id in self._neighbor_manager.known_neighbors:
            if neighbor_id in self._neighbor_metadata:
                neighbor = self._neighbor_metadata[neighbor_id]
                if self._my_state.state_differs(
                        distributed=self._distributed, other=neighbor.current_state):
                    positive_decisions.append(neighbor_id)
            else:
                positive_decisions.append(neighbor_id)
        if positive_decisions:
            log.info('node %s (%x) is unicasting now without decisions to %s.',
                     self._my_id, id(self), positive_decisions)
            self._neighbor_manager.send(positive_decisions, self._my_state.encode())
            self._my_state_metadata.sent_current_state()
            self._record_current_state_sent(positive_decisions)

    def _send_broadcast_decisions(self, decisions: List[Tuple[bool, NodeID]]) -> None:
        if any(decision for decision, _ in decisions):
            log.info('node %s (%x) is broadcasting now due to at least 1 decision.',
                     self._my_id, id(self))
            self._send()

    def _send_unicast_decisions(self, decisions: List[Tuple[bool, NodeID]]) -> None:
        assert self._my_state is not None
        assert self._neighbor_manager is not None
        if positive_decisions := [
            neighbor_id for decision, neighbor_id in decisions if decision
        ]:
            log.info('node %s (%x) is unicasting to %s.',
                     self._my_id, id(self), positive_decisions)
            self._neighbor_manager.send(positive_decisions, self._my_state.encode())
            self._my_state_metadata.sent_current_state()
            self._record_current_state_sent(positive_decisions)

    def _conditionally_send(self):  # pylint: disable=too-many-branches
        '''Send the model to neighbors if the deciders say so.

        Must be called with the model lock held.
        '''
        # Bootstrap neighbor discovery.
        state_changed = self._my_state.state_differs(
            distributed=self._distributed,
            other=self._my_state_metadata.last_state_sent)

        if self._my_state_metadata.last_state_sent is None:
            log.info('node %s (%x) is sending for the first time.', self._my_id, id(self))
            self._send()
            return

        self._changed_on_fit = state_changed

        if self._distributed.no_deciders:
            log.info('node %s (%x) is sending without deciders.', self._my_id, id(self))
            if (self._distributed.communicator.strategy
                    == CommunicatorConfig.Strategies.BROADCAST.value):
                log.info('node %s (%x) is broadcasting without decisions.', self._my_id, id(self))
                self._send_broadcast_no_decisions(state_changed=state_changed)
            elif (self._distributed.communicator.strategy
                    == CommunicatorConfig.Strategies.UNICAST.value):
                log.info('node %s (%x) is unicasting without decisions.', self._my_id, id(self))
                self._send_unicast_no_decisions()
            else:
                raise NotImplementedError(
                    f'Unknown strategy: {self._distributed.communicator.strategy}')
        else:
            log.info('node %s (%x) has deciders.', self._my_id, id(self))
            # We have deciders.
            decisions = []
            for neighbor_id in self._neighbor_manager.known_neighbors:
                if neighbor_id not in self._neighbor_metadata:
                    # TODO(piggy): Enhanced deciders should use this information.
                    decisions.append((True, neighbor_id))
                    continue
                decision = self.decide(neighbor_id=neighbor_id)
                decisions.append((decision, neighbor_id))

            log.info('node %s (%x) has decisions: %s', self._my_id, id(self), decisions)
            # Broadcast if any decider says to.
            if (self._distributed.communicator.strategy
                    == CommunicatorConfig.Strategies.BROADCAST.value):
                log.info('node %s (%x) is broadcasting with decisions.', self._my_id, id(self))
                self._send_broadcast_decisions(decisions=decisions)

            # Unicast to neighbors if a decider says to.
            elif (self._distributed.communicator.strategy
                    == CommunicatorConfig.Strategies.UNICAST.value):
                log.info('node %s (%x) is unicasting with decisions.', self._my_id, id(self))
                self._send_unicast_decisions(decisions=decisions)
            else:
                raise NotImplementedError(
                    'Have deciders, but unknown strategy: '
                    f'{self._distributed.communicator.strategy}')
        self._detect_convergence(state_changed=state_changed)

    @abc.abstractmethod
    def _fit(self, dataset: Optional[Dataset], **kwargs) -> None:
        '''Fit the model to the data. This is the actual implementation of fit.'''

    def predict(self, dataset: Optional[Dataset], **kwargs) -> Optional[Dataset]:
        '''Apply model to input dataset to create output.

        This handles lock acquisition and generalized distributed work.
        '''

        log.info('node %s (%x) is predicting (%s).',
                 self._my_id,
                 id(self),
                 'no data' if dataset is None else 'with data')

        if not self._trained and not self._training:
            # As a convenience, wait 2 polling intervals to see if there is a fit
            #   request on the way.  We want to wait just over 1 polling
            #   interval, since that's how long it could take for a pending
            #   fit request to start training.
            time.sleep(self._distributed.polling_interval * 2)

            if not self._trained and not self._training:
                raise UntrainedError(
                    f'Algorithm "{self.catalog_name}" needs to be fit before it can predict')

        assert self._lock_predict_state is not None

        # On the first fit, we spin until we get a predict state.
        ok_to_proceed = False
        timeout = time.monotonic_ns() + self._distributed.predict_wait_for_fit_timeout_ns
        while not ok_to_proceed:
            if time.monotonic_ns() > timeout:
                raise UntrainedError(
                    f'Algorithm "{self.catalog_name}" ran out of time in predict waiting for fit')
            with self._lock_predict_state:
                if self._predict_state is not None:
                    ok_to_proceed = True
            time.sleep(self._distributed.polling_interval)

        with self._lock_predict_state:
            if dataset is not None:
                if self._is_empty(dataset):
                    log.info('Node %s encountered empty or malformed dataset on predict: \n%s\n',
                             self._my_id, dataset)
                else:
                    dataset = dataset.sorted_columns()
            retval = self._predict(dataset=dataset, my_state=self._predict_state, **kwargs)
            self._predict_counter += 1

        log.info('Node %s predictions: \n%s', self._my_id, retval)

        return retval

    @abc.abstractmethod
    def _predict(self, dataset: Optional[Dataset], **kwargs) -> Optional[Dataset]:
        '''Apply model to input dataset to create output.

        This handles model-speific work and does not handle general tasks.q
        '''



'''Tests for the Aggregator Base class implementations'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...instantiator.executable_pipeline import PipelineResults
from ...metrics.impl.metric import MetricStub
from ...ranker.ranker import Ranking, Rankings, RankingStub
from .aggregator import Aggregator

# pylint: disable=missing-function-docstring,duplicate-code
# pylint: disable=missing-class-docstring
# pylint: disable=too-few-public-methods


class ExampleAggregator(Aggregator):
    _name = 'TestAggregator'

    def aggregate(self, rankings: Rankings, all_scores: bool = False) -> Ranking:
        return RankingStub(MetricStub(), PipelineResults())


def test_name():
    dut = ExampleAggregator()
    assert dut.name == 'testaggregator'
    got = dut.aggregate({})
    assert got.metric.name == 'stub_metric'
